import numpy as np
import pandas as pd
import pytest

from woodwork.accessor_utils import (
    _is_dask_dataframe,
    _is_dask_series,
    _is_dataframe,
    _is_koalas_dataframe,
    _is_koalas_series,
    _is_series,
    get_invalid_schema_message,
    init_series,
    is_schema_valid
)
from woodwork.exceptions import TypeConversionError
from woodwork.logical_types import Categorical, Datetime, NaturalLanguage


def test_init_series_valid_conversion_specified_ltype(sample_series):
    if _is_koalas_series(sample_series):
        sample_series = sample_series.astype('str')
    else:
        sample_series = sample_series.astype('object')

    series = init_series(sample_series, logical_type='categorical')
    assert series is not sample_series
    correct_dtype = Categorical._get_valid_dtype(type(sample_series))
    assert series.dtype == correct_dtype
    assert isinstance(series.ww.logical_type, Categorical)
    assert series.ww.semantic_tags == {'category'}

    series = init_series(sample_series, logical_type='natural_language')
    assert series is not sample_series
    correct_dtype = NaturalLanguage._get_valid_dtype(type(sample_series))
    assert series.dtype == correct_dtype
    assert isinstance(series.ww.logical_type, NaturalLanguage)
    assert series.ww.semantic_tags == set()


def test_init_series_with_pd_extension_array():
    extension_categories = pd.Categorical([1, 2, 3])
    series = init_series(extension_categories)
    pd_reference_series = init_series(pd.Series([1, 2, 3], dtype='category'))
    assert series.equals(pd_reference_series)
    assert series.ww.logical_type == pd_reference_series.ww.logical_type
    assert series.ww.semantic_tags == pd_reference_series.ww.semantic_tags

    extension_ints = pd.array(np.array([1, 2, 3, 4], dtype="int64"))
    series = init_series(extension_ints)
    pd_reference_series = init_series(pd.Series([1, 2, 3, 4], dtype='Int64'))
    assert series.equals(pd_reference_series)
    assert series.ww.logical_type == pd_reference_series.ww.logical_type
    assert series.ww.semantic_tags == pd_reference_series.ww.semantic_tags


def test_init_series_with_invalid_type(sample_df):
    inputs = [sample_df, 1, "string", None]
    for input_ in inputs:
        error_message = f'Input must be of series type. The current input is of type {type(input_)}'
        with pytest.raises(TypeError, match=error_message):
            init_series(input_)


def test_init_series_with_np_array(sample_series_pandas):
    series = init_series(sample_series_pandas.to_numpy())
    series2 = init_series(sample_series_pandas)  # Sample series panda contains ['a','b','c','a']
    assert series.equals(series2)
    assert series.ww.logical_type == series2.ww.logical_type
    assert series.ww.semantic_tags == series2.ww.semantic_tags


def test_init_series_with_multidimensional_np_array():
    input_ = np.array([['a', 'b'], ['a', 'b']])
    error_message = f'np.ndarray input must be 1 dimensional. Current np.ndarray is {input_.ndim} dimensional'
    with pytest.raises(ValueError, match=error_message):
        init_series(input_)


def test_init_series_valid_conversion_inferred_ltype(sample_series):
    if _is_koalas_series(sample_series):
        sample_series = sample_series.astype('str')
    else:
        sample_series = sample_series.astype('object')

    series = init_series(sample_series)
    assert series is not sample_series
    correct_dtype = Categorical._get_valid_dtype(type(sample_series))
    assert series.dtype == correct_dtype
    assert isinstance(series.ww.logical_type, Categorical)
    assert series.ww.semantic_tags == {'category'}


def test_init_series_with_datetime(sample_datetime_series):
    series = init_series(sample_datetime_series, logical_type='datetime')
    assert series.dtype == 'datetime64[ns]'
    assert isinstance(series.ww.logical_type, Datetime)


def test_init_series_all_parameters(sample_series):
    if _is_koalas_series(sample_series):
        sample_series = sample_series.astype('str')
    else:
        sample_series = sample_series.astype('object')

    metadata = {'meta_key': 'meta_value'}
    description = 'custom description'
    origin = 'base'
    series = init_series(sample_series,
                         logical_type='categorical',
                         semantic_tags=['custom_tag'],
                         metadata=metadata,
                         description=description,
                         origin=origin,
                         use_standard_tags=False)
    assert series is not sample_series
    correct_dtype = Categorical._get_valid_dtype(type(sample_series))
    assert series.dtype == correct_dtype
    assert isinstance(series.ww.logical_type, Categorical)
    assert series.ww.semantic_tags == {'custom_tag'}
    assert series.ww.metadata == metadata
    assert series.ww.description == description
    assert series.ww.origin == origin


def test_init_series_error_on_invalid_conversion(sample_series):
    if _is_dask_series(sample_series):
        pytest.xfail('Dask type conversion with astype does not fail until compute is called')
    if _is_koalas_series(sample_series):
        pytest.xfail('Koalas allows this conversion, filling values it cannot convert with NaN '
                     'and converting dtype to float.')

    error_message = "Error converting datatype for sample_series from type category to type Int64. " \
        "Please confirm the underlying data is consistent with logical type IntegerNullable."
    with pytest.raises(TypeConversionError, match=error_message):
        init_series(sample_series, logical_type='integer_nullable')


def test_is_series(sample_df):
    assert _is_series(sample_df['id'])
    assert not _is_series(sample_df)


def test_is_dataframe(sample_df):
    assert _is_dataframe(sample_df)
    assert not _is_dataframe(sample_df['id'])


def test_get_invalid_schema_message(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(name='test_schema', index='id', logical_types={'id': 'Double', 'full_name': 'PersonFullName'})
    schema = schema_df.ww.schema

    assert get_invalid_schema_message(schema_df, schema) is None
    assert (get_invalid_schema_message(sample_df, schema) ==
            'dtype mismatch for column id between DataFrame dtype, int64, and Double dtype, float64')

    sampled_df = schema_df.sample(frac=0.3)
    assert get_invalid_schema_message(sampled_df, schema) is None

    dropped_df = schema_df.drop('id', axis=1)
    assert (get_invalid_schema_message(dropped_df, schema) ==
            "The following columns in the typing information were missing from the DataFrame: {'id'}")

    renamed_df = schema_df.rename(columns={'id': 'new_col'})
    assert (get_invalid_schema_message(renamed_df, schema) ==
            "The following columns in the DataFrame were missing from the typing information: {'new_col'}")


def test_get_invalid_schema_message_dtype_mismatch(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(logical_types={'age': 'Categorical', 'full_name': 'PersonFullName'})
    schema = schema_df.ww.schema

    incorrect_int_dtype_df = schema_df.ww.astype({'id': 'Int64'})
    incorrect_bool_dtype_df = schema_df.ww.astype({'is_registered': 'Int64'})

    assert (get_invalid_schema_message(incorrect_int_dtype_df, schema) ==
            'dtype mismatch for column id between DataFrame dtype, Int64, and Integer dtype, int64')
    assert (get_invalid_schema_message(incorrect_bool_dtype_df, schema) ==
            'dtype mismatch for column is_registered between DataFrame dtype, Int64, and BooleanNullable dtype, boolean')

    # Koalas backup dtypes make these checks not relevant
    if not _is_koalas_dataframe(sample_df):
        incorrect_str_dtype_df = schema_df.ww.astype({'full_name': 'object'})  # wont work for koalas
        incorrect_categorical_dtype_df = schema_df.ww.astype({'age': 'string'})  # wont work for koalas
        assert (get_invalid_schema_message(incorrect_str_dtype_df, schema) ==
                'dtype mismatch for column full_name between DataFrame dtype, object, and PersonFullName dtype, string')
        assert (get_invalid_schema_message(incorrect_categorical_dtype_df, schema) ==
                'dtype mismatch for column age between DataFrame dtype, string, and Categorical dtype, category')


def test_get_invalid_schema_message_index_checks(sample_df):
    if not isinstance(sample_df, pd.DataFrame):
        pytest.xfail('Index validation not performed for Dask or Koalas DataFrames')

    schema_df = sample_df.copy()
    schema_df.ww.init(name='test_schema', index='id', logical_types={'id': 'Double', 'full_name': 'PersonFullName'})
    schema = schema_df.ww.schema

    different_underlying_index_df = schema_df.copy()
    different_underlying_index_df['id'] = pd.Series([9, 8, 7, 6], dtype='float64')
    assert (get_invalid_schema_message(different_underlying_index_df, schema) ==
            "Index mismatch between DataFrame and typing information")

    not_unique_df = schema_df.replace({3: 1})
    not_unique_df.index = not_unique_df['id']
    not_unique_df.index.name = None
    assert get_invalid_schema_message(not_unique_df, schema) == 'Index column is not unique'

    nan_df = pd.DataFrame({'id': pd.Series([5, 4, None, 2], dtype='float64'), 'col': pd.Series(['b', 'b', 'b', 'd'], dtype='category')})
    nan_df.ww.init(index='id')
    nan_df_schema = nan_df.ww.schema

    assert get_invalid_schema_message(nan_df, nan_df_schema) is None


def test_is_schema_valid_true(sample_df):
    sample_df.ww.init(index='id', logical_types={'phone_number': 'Categorical'})
    copy_df = sample_df.copy()

    assert copy_df.ww.schema is None
    assert is_schema_valid(copy_df, sample_df.ww.schema)


def test_is_schema_valid_false(sample_df):
    sample_df.ww.init()
    schema = sample_df.ww.schema

    invalid_dtype_df = sample_df.astype({'age': 'float64'})
    assert not is_schema_valid(invalid_dtype_df, schema)

    missing_col_df = sample_df.drop(columns={'is_registered'})
    assert not is_schema_valid(missing_col_df, schema)


def test_is_dask_dataframe(sample_df_dask):
    assert _is_dask_dataframe(sample_df_dask)
    assert not _is_dask_dataframe(pd.DataFrame())


def test_is_dask_series(sample_series_dask):
    assert _is_dask_series(sample_series_dask)
    assert not _is_dask_series(pd.Series())


def test_is_koalas_dataframe(sample_df_koalas):
    assert _is_koalas_dataframe(sample_df_koalas)
    assert not _is_dask_dataframe(pd.DataFrame())


def test_is_koalas_series(sample_series_koalas):
    assert _is_koalas_series(sample_series_koalas)
    assert not _is_dask_series(pd.Series())



import re
from datetime import datetime
from inspect import isclass

import numpy as np
import pandas as pd
import pytest

from woodwork.accessor_utils import _is_koalas_dataframe, init_series
from woodwork.logical_types import (
    URL,
    Age,
    AgeNullable,
    Boolean,
    BooleanNullable,
    Categorical,
    CountryCode,
    Datetime,
    Double,
    EmailAddress,
    Filepath,
    Integer,
    IntegerNullable,
    IPAddress,
    LatLong,
    NaturalLanguage,
    Ordinal,
    PersonFullName,
    PhoneNumber,
    PostalCode,
    SubRegionCode,
    Timedelta
)
from woodwork.statistics_utils import (
    _get_describe_dict,
    _get_histogram_values,
    _get_mode,
    _get_numeric_value_counts_in_range,
    _get_recent_value_counts,
    _get_top_values_categorical,
    _make_categorical_for_mutual_info,
    _replace_nans_for_mutual_info
)
from woodwork.tests.testing_utils import (
    check_empty_box_plot_dict,
    mi_between_cols,
    to_pandas
)


def test_get_mode():
    series_list = [
        pd.Series([1, 2, 3, 4, 2, 2, 3]),
        pd.Series(['a', 'b', 'b', 'c', 'b']),
        pd.Series([3, 2, 3, 2]),
        pd.Series([np.nan, np.nan, np.nan]),
        pd.Series([pd.NA, pd.NA, pd.NA]),
        pd.Series([1, 2, np.nan, 2, np.nan, 3, 2]),
        pd.Series([1, 2, pd.NA, 2, pd.NA, 3, 2])
    ]

    answer_list = [2, 'b', 2, None, None, 2, 2]

    for series, answer in zip(series_list, answer_list):
        mode = _get_mode(series)
        if answer is None:
            assert mode is None
        else:
            assert mode == answer


def test_accessor_replace_nans_for_mutual_info():
    df_nans = pd.DataFrame({
        'ints': pd.Series([2, pd.NA, 5, 2], dtype='Int64'),
        'floats': pd.Series([3.3, None, 2.3, 1.3]),
        'bools': pd.Series([True, None, True, False]),
        'bools_pdna': pd.Series([True, pd.NA, True, False], dtype='boolean'),
        'int_to_cat_nan': pd.Series([1, np.nan, 3, 1], dtype='category'),
        'str': pd.Series(['test', np.nan, 'test2', 'test']),
        'str_no_nan': pd.Series(['test', 'test2', 'test2', 'test']),
        'dates': pd.Series(['2020-01-01', None, '2020-01-02', '2020-01-03'])
    })
    df_nans.ww.init(logical_types={"bools": "Categorical", "str":
                                   "Categorical", "str_no_nan": "Categorical"})
    formatted_df = _replace_nans_for_mutual_info(df_nans.ww.schema, df_nans.copy())

    assert isinstance(formatted_df, pd.DataFrame)

    assert formatted_df['ints'].equals(pd.Series([2, 3, 5, 2], dtype='Int64'))
    assert formatted_df['floats'].equals(pd.Series([3.3, 2.3, 2.3, 1.3], dtype='float'))
    assert formatted_df['bools'].equals(pd.Series([True, True, True, False], dtype='category'))
    assert formatted_df['bools_pdna'].equals(pd.Series([True, True, True, False], dtype='boolean'))
    assert formatted_df['int_to_cat_nan'].equals(pd.Series([1, 1, 3, 1], dtype='category'))
    assert formatted_df['str'].equals(pd.Series(['test', 'test', 'test2', 'test'], dtype='category'))
    assert formatted_df['str_no_nan'].equals(pd.Series(['test', 'test2', 'test2', 'test'], dtype='category'))
    assert formatted_df['dates'].equals(pd.Series(['2020-01-01', '2020-01-02', '2020-01-02', '2020-01-03'], dtype='datetime64[ns]'))


def test_accessor_make_categorical_for_mutual_info():
    df = pd.DataFrame({
        'ints1': pd.Series([1, 2, 3, 2]),
        'ints2': pd.Series([1, 100, 1, 100]),
        'ints3': pd.Series([1, 2, 3, 2], dtype='Int64'),
        'bools': pd.Series([True, False, True, False]),
        'booleans': pd.Series([True, False, True, False], dtype='boolean'),
        'categories': pd.Series(['test', 'test2', 'test2', 'test']),
        'dates': pd.Series(['2020-01-01', '2019-01-02', '2020-08-03', '1997-01-04'])
    })
    df.ww.init()
    formatted_num_bins_df = _make_categorical_for_mutual_info(df.ww.schema, df.copy(), num_bins=4)

    assert isinstance(formatted_num_bins_df, pd.DataFrame)

    assert formatted_num_bins_df['ints1'].equals(pd.Series([0, 1, 3, 1], dtype='int8'))
    assert formatted_num_bins_df['ints2'].equals(pd.Series([0, 1, 0, 1], dtype='int8'))
    assert formatted_num_bins_df['ints3'].equals(pd.Series([0, 1, 3, 1], dtype='int8'))
    assert formatted_num_bins_df['bools'].equals(pd.Series([1, 0, 1, 0], dtype='int8'))
    assert formatted_num_bins_df['booleans'].equals(pd.Series([1, 0, 1, 0], dtype='int8'))
    assert formatted_num_bins_df['categories'].equals(pd.Series([0, 1, 1, 0], dtype='int8'))
    assert formatted_num_bins_df['dates'].equals(pd.Series([2, 1, 3, 0], dtype='int8'))


def test_mutual_info_same(df_same_mi):
    df_same_mi.ww.init(logical_types={'nans': Categorical()})

    mi = df_same_mi.ww.mutual_information()

    cols_used = set(np.unique(mi[['column_1', 'column_2']].values))
    assert 'nans' not in cols_used
    assert 'nat_lang' not in cols_used
    assert mi.shape[0] == 1
    assert mi_between_cols('floats', 'ints', mi) == 1.0


def test_mutual_info(df_mi):
    df_mi.ww.init(logical_types={'dates': Datetime(datetime_format='%Y-%m-%d')})
    original_df = df_mi.copy()
    mi = df_mi.ww.mutual_information()
    assert mi.shape[0] == 10

    np.testing.assert_almost_equal(mi_between_cols('ints', 'bools', mi), 1.0, 3)
    np.testing.assert_almost_equal(mi_between_cols('ints', 'strs', mi), 0.0, 3)
    np.testing.assert_almost_equal(mi_between_cols('strs', 'bools', mi), 0, 3)
    np.testing.assert_almost_equal(mi_between_cols('dates', 'ints', mi), 0.274, 3)
    np.testing.assert_almost_equal(mi_between_cols('dates', 'bools', mi), 0.274, 3)

    mi_many_rows = df_mi.ww.mutual_information(nrows=100000)
    pd.testing.assert_frame_equal(mi, mi_many_rows)

    mi = df_mi.ww.mutual_information(nrows=1)
    assert mi.shape[0] == 10
    assert (mi['mutual_info'] == 1.0).all()

    mi = df_mi.ww.mutual_information(num_bins=2)
    assert mi.shape[0] == 10
    np.testing.assert_almost_equal(mi_between_cols('bools', 'ints', mi), 0.0, 3)
    np.testing.assert_almost_equal(mi_between_cols('strs', 'ints', mi), 1.0, 3)
    np.testing.assert_almost_equal(mi_between_cols('bools', 'strs', mi), 0, 3)
    np.testing.assert_almost_equal(mi_between_cols('dates', 'strs', mi), 1.0, 3)
    np.testing.assert_almost_equal(mi_between_cols('dates', 'ints', mi), 1.0, 3)

    # Confirm that none of this changed the underlying df
    pd.testing.assert_frame_equal(to_pandas(df_mi), to_pandas(original_df))


def test_mutual_info_on_index(sample_df):
    sample_df.ww.init(index='id')
    mi = sample_df.ww.mutual_information()

    assert not ('id' in mi['column_1'].values or 'id' in mi['column_2'].values)

    mi = sample_df.ww.mutual_information(include_index=True)
    assert 'id' in mi['column_1'].values or 'id' in mi['column_2'].values


def test_mutual_info_returns_empty_df_properly(sample_df):
    schema_df = sample_df[['id', 'age']]
    schema_df.ww.init(index='id')

    mi = schema_df.ww.mutual_information()
    assert mi.empty


def test_mutual_info_sort(df_mi):
    df_mi.ww.init()
    mi = df_mi.ww.mutual_information()

    for i in range(len(mi['mutual_info']) - 1):
        assert mi['mutual_info'].iloc[i] >= mi['mutual_info'].iloc[i + 1]


def test_mutual_info_dict(df_mi):
    df_mi.ww.init()
    mi_dict = df_mi.ww.mutual_information_dict()
    mi = df_mi.ww.mutual_information()

    pd.testing.assert_frame_equal(pd.DataFrame(mi_dict), mi)


def test_mutual_info_unique_cols(df_mi_unique):
    df_mi_unique.ww.init()
    mi = df_mi_unique.ww.mutual_information()

    cols_used = set(np.unique(mi[['column_1', 'column_2']].values))
    assert 'unique' in cols_used
    assert 'unique_with_one_nan' in cols_used
    assert 'unique_with_nans' in cols_used
    assert 'ints' in cols_used


def test_mutual_info_callback(df_mi):
    df_mi.ww.init(logical_types={'dates': Datetime(datetime_format='%Y-%m-%d')})

    class MockCallback:
        def __init__(self):
            self.progress_history = []
            self.total_update = 0
            self.total_elapsed_time = 0

        def __call__(self, update, progress, total, unit, time_elapsed):
            self.total_update += update
            self.total = total
            self.progress_history.append(progress)
            self.unit = unit
            self.total_elapsed_time = time_elapsed

    mock_callback = MockCallback()

    df_mi.ww.mutual_information(callback=mock_callback)

    # Should be 18 total calls
    assert len(mock_callback.progress_history) == 18

    assert mock_callback.unit == 'calculations'
    # First call should be 1 of 26 calculations complete
    assert mock_callback.progress_history[0] == 1
    # After second call should be 6 of 26 units complete
    assert mock_callback.progress_history[1] == 6

    # Should be 26 calculations at end with a positive elapsed time
    assert mock_callback.total == 26
    assert mock_callback.total_update == 26
    assert mock_callback.progress_history[-1] == 26
    assert mock_callback.total_elapsed_time > 0


def test_get_describe_dict(describe_df):
    describe_df.ww.init(index='index_col')

    stats_dict = _get_describe_dict(describe_df)
    stats_dict_to_df = pd.DataFrame(stats_dict)
    for extra in ['histogram', 'top_values', 'recent_values']:
        assert extra not in stats_dict_to_df.index

    index_order = ['physical_type',
                   'logical_type',
                   'semantic_tags',
                   'count',
                   'nunique',
                   'nan_count',
                   'mean',
                   'mode',
                   'std',
                   'min',
                   'first_quartile',
                   'second_quartile',
                   'third_quartile',
                   'max',
                   'num_true',
                   'num_false']

    stats_dict_to_df = stats_dict_to_df.reindex(index_order)
    stats_df = describe_df.ww.describe()
    pd.testing.assert_frame_equal(stats_df, stats_dict_to_df)


def test_describe_does_not_include_index(describe_df):
    describe_df.ww.init(index='index_col')
    stats_df = describe_df.ww.describe()
    assert 'index_col' not in stats_df.columns


def test_describe_accessor_method(describe_df):
    categorical_ltypes = [Categorical,
                          CountryCode,
                          Ordinal(order=('yellow', 'red', 'blue')),
                          PostalCode,
                          SubRegionCode]
    boolean_ltypes = [BooleanNullable]
    non_nullable_boolean_ltypes = [Boolean]
    datetime_ltypes = [Datetime]
    formatted_datetime_ltypes = [Datetime(datetime_format='%Y~%m~%d')]
    timedelta_ltypes = [Timedelta]
    nullable_numeric_ltypes = [Double, IntegerNullable, AgeNullable]
    non_nullable_numeric_ltypes = [Integer, Age]
    natural_language_ltypes = [EmailAddress, Filepath, PersonFullName, IPAddress,
                               PhoneNumber, URL]
    latlong_ltypes = [LatLong]

    expected_index = ['physical_type',
                      'logical_type',
                      'semantic_tags',
                      'count',
                      'nunique',
                      'nan_count',
                      'mean',
                      'mode',
                      'std',
                      'min',
                      'first_quartile',
                      'second_quartile',
                      'third_quartile',
                      'max',
                      'num_true',
                      'num_false']

    # Test categorical columns
    category_data = describe_df[['category_col']]
    if _is_koalas_dataframe(category_data):
        expected_dtype = 'string'
    else:
        expected_dtype = 'category'

    for ltype in categorical_ltypes:
        if isclass(ltype):
            ltype = ltype()

        expected_vals = pd.Series({
            'physical_type': expected_dtype,
            'logical_type': ltype,
            'semantic_tags': {'category', 'custom_tag'},
            'count': 7,
            'nunique': 3,
            'nan_count': 1,
            'mode': 'red'}, name='category_col')
        category_data.ww.init(logical_types={'category_col': ltype}, semantic_tags={'category_col': 'custom_tag'})
        stats_df = category_data.ww.describe()
        assert isinstance(stats_df, pd.DataFrame)
        assert set(stats_df.columns) == {'category_col'}
        assert stats_df.index.tolist() == expected_index
        assert expected_vals.equals(stats_df['category_col'].dropna())

    # Test nullable boolean columns
    boolean_data = describe_df[['boolean_col']]
    for ltype in boolean_ltypes:
        expected_dtype = ltype.primary_dtype
        expected_vals = pd.Series({
            'physical_type': expected_dtype,
            'logical_type': ltype(),
            'semantic_tags': {'custom_tag'},
            'count': 7,
            'nan_count': 1,
            'mode': True,
            'num_true': 4,
            'num_false': 3}, name='boolean_col')
        boolean_data.ww.init(logical_types={'boolean_col': ltype}, semantic_tags={'boolean_col': 'custom_tag'})
        stats_df = boolean_data.ww.describe()
        assert isinstance(stats_df, pd.DataFrame)
        assert set(stats_df.columns) == {'boolean_col'}
        assert stats_df.index.tolist() == expected_index
        assert expected_vals.equals(stats_df['boolean_col'].dropna())

    # Test non-nullable boolean columns
    boolean_data = describe_df[['boolean_col']].fillna(True)
    for ltype in non_nullable_boolean_ltypes:
        expected_dtype = ltype.primary_dtype
        expected_vals = pd.Series({
            'physical_type': expected_dtype,
            'logical_type': ltype(),
            'semantic_tags': {'custom_tag'},
            'count': 8,
            'nan_count': 0,
            'mode': True,
            'num_true': 5,
            'num_false': 3}, name='boolean_col')
        boolean_data.ww.init(logical_types={'boolean_col': ltype}, semantic_tags={'boolean_col': 'custom_tag'})
        stats_df = boolean_data.ww.describe()
        assert isinstance(stats_df, pd.DataFrame)
        assert set(stats_df.columns) == {'boolean_col'}
        assert stats_df.index.tolist() == expected_index
        assert expected_vals.equals(stats_df['boolean_col'].dropna())

    # Test datetime columns
    datetime_data = describe_df[['datetime_col']]
    for ltype in datetime_ltypes:
        expected_vals = pd.Series({
            'physical_type': ltype.primary_dtype,
            'logical_type': ltype(),
            'semantic_tags': {'custom_tag'},
            'count': 7,
            'nunique': 6,
            'nan_count': 1,
            'mean': pd.Timestamp('2020-01-19 09:25:42.857142784'),
            'mode': pd.Timestamp('2020-02-01 00:00:00'),
            'min': pd.Timestamp('2020-01-01 00:00:00'),
            'max': pd.Timestamp('2020-02-02 18:00:00')}, name='datetime_col')
        datetime_data.ww.init(logical_types={'datetime_col': ltype}, semantic_tags={'datetime_col': 'custom_tag'})
        stats_df = datetime_data.ww.describe()
        assert isinstance(stats_df, pd.DataFrame)
        assert set(stats_df.columns) == {'datetime_col'}
        assert stats_df.index.tolist() == expected_index
        assert expected_vals.equals(stats_df['datetime_col'].dropna())

    # Test formatted datetime columns
    formatted_datetime_data = describe_df[['formatted_datetime_col']]
    for ltype in formatted_datetime_ltypes:
        converted_to_datetime = pd.to_datetime(['2020-01-01',
                                                '2020-02-01',
                                                '2020-03-01',
                                                '2020-02-02',
                                                '2020-03-02',
                                                pd.NaT,
                                                '2020-02-01',
                                                '2020-01-02'])
        expected_vals = pd.Series({
            'physical_type': ltype.primary_dtype,
            'logical_type': ltype,
            'semantic_tags': {'custom_tag'},
            'count': 7,
            'nunique': 6,
            'nan_count': 1,
            'mean': converted_to_datetime.mean(),
            'mode': pd.to_datetime('2020-02-01'),
            'min': converted_to_datetime.min(),
            'max': converted_to_datetime.max()}, name='formatted_datetime_col')
        formatted_datetime_data.ww.init(
            logical_types={'formatted_datetime_col': ltype},
            semantic_tags={'formatted_datetime_col': 'custom_tag'})
        stats_df = formatted_datetime_data.ww.describe()
        assert isinstance(stats_df, pd.DataFrame)
        assert set(stats_df.columns) == {'formatted_datetime_col'}
        assert stats_df.index.tolist() == expected_index
        assert expected_vals.equals(stats_df['formatted_datetime_col'].dropna())

    # Test timedelta columns - Skip for Koalas
    if not _is_koalas_dataframe(describe_df):
        timedelta_data = describe_df['timedelta_col']
        for ltype in timedelta_ltypes:
            expected_vals = pd.Series({
                'physical_type': ltype.primary_dtype,
                'logical_type': ltype(),
                'semantic_tags': {'custom_tag'},
                'count': 7,
                'nan_count': 1,
                'mode': pd.Timedelta('31days')}, name='col')
            df = pd.DataFrame({'col': timedelta_data})
            df.ww.init(logical_types={'col': ltype}, semantic_tags={'col': 'custom_tag'})
            stats_df = df.ww.describe()
            assert isinstance(stats_df, pd.DataFrame)
            assert set(stats_df.columns) == {'col'}
            assert stats_df.index.tolist() == expected_index
            assert expected_vals.equals(stats_df['col'].dropna())

    # Test numeric columns with nullable ltypes
    numeric_data = describe_df[['numeric_col']]
    for ltype in nullable_numeric_ltypes:
        expected_vals = pd.Series({
            'physical_type': ltype.primary_dtype,
            'logical_type': ltype(),
            'semantic_tags': {'numeric', 'custom_tag'},
            'count': 7,
            'nunique': 6,
            'nan_count': 1,
            'mean': 20.857142857142858,
            'mode': 10,
            'std': 18.27957486220227,
            'min': 1,
            'first_quartile': 10,
            'second_quartile': 17,
            'third_quartile': 26,
            'max': 56}, name='numeric_col')
        numeric_data.ww.init(logical_types={'numeric_col': ltype}, semantic_tags={'numeric_col': 'custom_tag'})
        stats_df = numeric_data.ww.describe()
        assert isinstance(stats_df, pd.DataFrame)
        assert set(stats_df.columns) == {'numeric_col'}
        assert stats_df.index.tolist() == expected_index
        assert expected_vals.equals(stats_df['numeric_col'].dropna())

    # Test numeric with non-nullable ltypes
    numeric_data = describe_df[['numeric_col']].fillna(0)
    for ltype in non_nullable_numeric_ltypes:
        expected_vals = pd.Series({
            'physical_type': ltype.primary_dtype,
            'logical_type': ltype(),
            'semantic_tags': {'numeric', 'custom_tag'},
            'count': 8,
            'nunique': 7,
            'nan_count': 0,
            'mean': 18.25,
            'mode': 10,
            'std': 18.460382289804137,
            'min': 0,
            'first_quartile': 7.75,
            'second_quartile': 13.5,
            'third_quartile': 23,
            'max': 56}, name='numeric_col')
        numeric_data.ww.init(logical_types={'numeric_col': ltype}, semantic_tags={'numeric_col': 'custom_tag'})
        stats_df = numeric_data.ww.describe()
        assert isinstance(stats_df, pd.DataFrame)
        assert set(stats_df.columns) == {'numeric_col'}
        assert stats_df.index.tolist() == expected_index
        assert expected_vals.equals(stats_df['numeric_col'].dropna())

    # Test natural language columns
    natural_language_data = describe_df[['natural_language_col']]
    expected_dtype = 'string'
    for ltype in natural_language_ltypes:
        expected_vals = pd.Series({
            'physical_type': expected_dtype,
            'logical_type': ltype(),
            'semantic_tags': {'custom_tag'},
            'count': 7,
            'nan_count': 1,
            'mode': 'Duplicate sentence.'}, name='natural_language_col')
        natural_language_data.ww.init(
            logical_types={'natural_language_col': ltype},
            semantic_tags={'natural_language_col': 'custom_tag'})
        stats_df = natural_language_data.ww.describe()
        assert isinstance(stats_df, pd.DataFrame)
        assert set(stats_df.columns) == {'natural_language_col'}
        assert stats_df.index.tolist() == expected_index
        assert expected_vals.equals(stats_df['natural_language_col'].dropna())

    # Test latlong columns
    latlong_data = describe_df[['latlong_col']]
    expected_dtype = 'object'
    for ltype in latlong_ltypes:
        mode = [0, 0] if _is_koalas_dataframe(describe_df) else (0, 0)
        expected_vals = pd.Series({
            'physical_type': expected_dtype,
            'logical_type': ltype(),
            'semantic_tags': {'custom_tag'},
            'count': 6,
            'nan_count': 2,
            'mode': mode}, name='latlong_col')
        latlong_data.ww.init(
            logical_types={'latlong_col': ltype},
            semantic_tags={'latlong_col': 'custom_tag'})
        stats_df = latlong_data.ww.describe()
        assert isinstance(stats_df, pd.DataFrame)
        assert set(stats_df.columns) == {'latlong_col'}
        assert stats_df.index.tolist() == expected_index
        assert expected_vals.equals(stats_df['latlong_col'].dropna())


def test_describe_with_improper_tags(describe_df):
    df = describe_df.copy()[['boolean_col', 'natural_language_col']]

    logical_types = {
        'boolean_col': Boolean,
        'natural_language_col': NaturalLanguage,
    }
    semantic_tags = {
        'boolean_col': 'category',
        'natural_language_col': 'numeric',
    }

    df.ww.init(logical_types=logical_types, semantic_tags=semantic_tags)
    stats_df = df.ww.describe()

    # Make sure boolean stats were computed with improper 'category' tag
    assert isinstance(stats_df['boolean_col']['logical_type'], Boolean)
    assert stats_df['boolean_col']['semantic_tags'] == {'category'}
    # Make sure numeric stats were not computed with improper 'numeric' tag
    assert stats_df['natural_language_col']['semantic_tags'] == {'numeric'}
    assert stats_df['natural_language_col'][['mean', 'std', 'min', 'max']].isnull().all()


def test_describe_with_no_semantic_tags(describe_df):
    df = describe_df.copy()[['category_col', 'numeric_col']]

    logical_types = {
        'category_col': Categorical,
        'numeric_col': IntegerNullable,
    }

    df.ww.init(logical_types=logical_types, use_standard_tags=False)
    stats_df = df.ww.describe()
    assert df.ww.semantic_tags['category_col'] == set()
    assert df.ww.semantic_tags['numeric_col'] == set()

    # Make sure category stats were computed
    assert stats_df['category_col']['semantic_tags'] == set()
    assert stats_df['category_col']['nunique'] == 3
    # Make sure numeric stats were computed
    assert stats_df['numeric_col']['semantic_tags'] == set()
    np.testing.assert_almost_equal(stats_df['numeric_col']['mean'], 20.85714, 5)


def test_describe_with_include(sample_df):
    semantic_tags = {
        'full_name': 'tag1',
        'email': ['tag2'],
        'age': ['numeric', 'age']
    }
    sample_df.ww.init(semantic_tags=semantic_tags)

    col_name_df = sample_df.ww.describe(include=['full_name'])
    assert col_name_df.shape == (16, 1)
    assert 'full_name', 'email' in col_name_df.columns

    semantic_tags_df = sample_df.ww.describe(['tag1', 'tag2'])
    assert 'full_name' in col_name_df.columns
    assert len(semantic_tags_df.columns) == 2

    logical_types_df = sample_df.ww.describe([Datetime, BooleanNullable])
    assert set(logical_types_df.columns) == {'signup_date', 'is_registered', 'datetime_with_NaT'}

    multi_params_df = sample_df.ww.describe(['age', 'tag1', Datetime])
    expected = ['full_name', 'age', 'signup_date', 'datetime_with_NaT']
    for col_name in expected:
        assert col_name in multi_params_df.columns
    multi_params_df['full_name'].equals(col_name_df['full_name'])
    multi_params_df['full_name'].equals(sample_df.ww.describe()['full_name'])


def test_pandas_nullable_integer_quantile_fix():
    """Should fail when https://github.com/pandas-dev/pandas/issues/42626 gets fixed"""
    if pd.__version__ not in ['1.3.0', '1.3.1']:  # pragma: no cover
        pytest.skip('Bug only exists on pandas version 1.3.0 and 1.3.1')
    series = pd.Series([1, 2, 3], dtype='Int64')
    error_message = "cannot safely cast non-equivalent object to int64"
    with pytest.raises(TypeError, match=error_message):
        series.quantile([.75])


def test_describe_with_no_match(sample_df):
    sample_df.ww.init()
    df = sample_df.ww.describe(include=['wrongname'])
    assert df.empty


def test_describe_callback(describe_df):
    describe_df.ww.init(index='index_col')

    class MockCallback:
        def __init__(self):
            self.progress_history = []
            self.total_update = 0
            self.total_elapsed_time = 0

        def __call__(self, update, progress, total, unit, time_elapsed):
            self.total_update += update
            self.total = total
            self.progress_history.append(progress)
            self.unit = unit
            self.total_elapsed_time = time_elapsed

    mock_callback = MockCallback()

    describe_df.ww.describe(callback=mock_callback)

    assert mock_callback.unit == 'calculations'
    # Koalas df does not have timedelta column
    if _is_koalas_dataframe(describe_df):
        ncalls = 9
    else:
        ncalls = 10

    assert len(mock_callback.progress_history) == ncalls

    # First call should be 1 unit complete
    assert mock_callback.progress_history[0] == 1
    # After second call should be 2 unit complete
    assert mock_callback.progress_history[1] == 2

    # Should be ncalls at end with a positive elapsed time
    assert mock_callback.total == ncalls
    assert mock_callback.total_update == ncalls
    assert mock_callback.progress_history[-1] == ncalls
    assert mock_callback.total_elapsed_time > 0


def test_describe_dict_extra_stats(describe_df):
    describe_df = describe_df.drop(columns=['boolean_col', 'natural_language_col', 'formatted_datetime_col',
                                            'timedelta_col', 'latlong_col'])
    describe_df['nullable_integer_col'] = describe_df['numeric_col']
    describe_df['integer_col'] = describe_df['numeric_col'].fillna(0)
    describe_df['small_range_col'] = describe_df['numeric_col'].fillna(0) // 10

    ltypes = {
        'category_col': 'Categorical',
        'datetime_col': 'Datetime',
        'numeric_col': 'Double',
        'nullable_integer_col': 'IntegerNullable',
        'integer_col': 'Integer',
        'small_range_col': 'Integer',
    }
    describe_df.ww.init(index='index_col', logical_types=ltypes)
    desc_dict = describe_df.ww.describe_dict(extra_stats=True)

    # category columns should have top_values
    assert isinstance(desc_dict['category_col']['top_values'], list)
    assert desc_dict['category_col'].get('histogram') is None
    assert desc_dict['category_col'].get('recent_values') is None

    # datetime columns should have recent_values
    assert isinstance(desc_dict['datetime_col']['recent_values'], list)
    assert desc_dict['datetime_col'].get('histogram') is None
    assert desc_dict['datetime_col'].get('top_values') is None

    # numeric columns should have histogram
    for col in ['numeric_col', 'nullable_integer_col', 'integer_col', 'small_range_col']:
        assert isinstance(desc_dict[col]['histogram'], list)
        assert desc_dict[col].get('recent_values') is None
        if col == 'small_range_col':
            # If values are in a narrow range, top values should be present
            assert isinstance(desc_dict[col]['top_values'], list)
        else:
            assert desc_dict[col].get('top_values') is None


def test_value_counts(categorical_df):
    logical_types = {
        'ints': IntegerNullable,
        'categories1': Categorical,
        'bools': Boolean,
        'categories2': Categorical,
        'categories3': Categorical,
    }
    categorical_df.ww.init(logical_types=logical_types)
    val_cts = categorical_df.ww.value_counts()
    for col in categorical_df.ww.columns:
        if col in ['ints', 'bools']:
            assert col not in val_cts
        else:
            assert col in val_cts

    expected_cat1 = [{'value': 200, 'count': 4}, {'value': 100, 'count': 3}, {'value': 1, 'count': 2}, {'value': 3, 'count': 1}]
    # Koalas converts numeric categories to strings, so we need to update the expected values for this
    # Koalas will result in `None` instead of `np.nan` in categorical columns
    if _is_koalas_dataframe(categorical_df):
        updated_results = []
        for items in expected_cat1:
            updated_results.append({k: (str(v) if k == 'value' else v) for k, v in items.items()})
        expected_cat1 = updated_results

    assert val_cts['categories1'] == expected_cat1
    assert val_cts['categories2'] == [{'value': np.nan, 'count': 6}, {'value': 'test', 'count': 3}, {'value': 'test2', 'count': 1}]
    assert val_cts['categories3'] == [{'value': np.nan, 'count': 7}, {'value': 'test', 'count': 3}]

    val_cts_descending = categorical_df.ww.value_counts(ascending=True)
    for col, vals in val_cts_descending.items():
        for i in range(len(vals)):
            assert vals[i]['count'] == val_cts[col][-i - 1]['count']

    val_cts_dropna = categorical_df.ww.value_counts(dropna=True)
    assert val_cts_dropna['categories3'] == [{'value': 'test', 'count': 3}]

    val_cts_2 = categorical_df.ww.value_counts(top_n=2)
    for col in val_cts_2:
        assert len(val_cts_2[col]) == 2


def test_datetime_get_recent_value_counts():
    times = pd.Series([
        datetime(2019, 2, 2, 1, 10, 0, 1),
        datetime(2019, 4, 2, 2, 20, 1, 0),
        datetime(2019, 3, 1, 3, 30, 1, 0),
        datetime(2019, 5, 1, 4, 40, 1, 0),
        datetime(2019, 1, 1, 5, 50, 1, 0),
        datetime(2019, 4, 2, 6, 10, 1, 0),
        datetime(2019, 4, 2, 7, 20, 1, 0),
        datetime(2019, 5, 1, 8, 30, 0, 0),
        pd.NaT,
    ])
    values = _get_recent_value_counts(times, num_x=3)
    expected_values = [
        {"value": datetime(2019, 4, 2).date(), "count": 3},
        {"value": datetime(2019, 5, 1).date(), "count": 2},
        {"value": datetime(2019, 3, 1).date(), "count": 1},
    ]
    assert values == expected_values


def test_numeric_histogram():
    column = pd.Series(np.random.randn(1000))
    column.append(pd.Series([np.nan, " ", "test"]))
    bins = 7
    values = _get_histogram_values(column, bins=bins)
    assert len(values) == bins
    total = 0
    for info in values:
        assert "bins" in info
        assert "frequency" in info
        freq = info["frequency"]
        total += freq
    assert total == 1000


@pytest.mark.parametrize(
    "input_series, expected",
    [
        (
            ["a", "b", "b", "c", "c", "c", np.nan],
            [
                {"value": "c", "count": 3},
                {"value": "b", "count": 2},
                {"value": "a", "count": 1},
            ],
        ),
        (
            [1, 2, 2, 3],
            [
                {"value": 2, "count": 2},
                {"value": 1, "count": 1},
                {"value": 3, "count": 1},
            ],
        ),
    ],
)
def test_get_top_values_categorical(input_series, expected):
    column = pd.Series(input_series)
    top_values = _get_top_values_categorical(column, 10)
    assert top_values == expected


@pytest.mark.parametrize(
    "input_series, expected",
    [
        (
            pd.Series([1, 2, 2, 3, 3, 3, pd.NA], dtype='Int64'),
            [
                {"value": 3, "count": 3},
                {"value": 2, "count": 2},
                {"value": 1, "count": 1},
                {"value": 0, "count": 0},
            ],
        ),
        (
            pd.Series([1, 2, 2, 3], dtype='int64'),
            [
                {"value": 2, "count": 2},
                {"value": 1, "count": 1},
                {"value": 3, "count": 1},
                {"value": 0, "count": 0},
            ],
        ),
    ],
)
def test_get_numeric_value_counts_in_range(input_series, expected):
    column = input_series
    top_values = _get_numeric_value_counts_in_range(column, range(4))
    assert top_values == expected


def test_box_plot_outliers(outliers_df):
    outliers_series = outliers_df['has_outliers']
    outliers_series.ww.init()

    no_outliers_series = outliers_df['no_outliers']
    no_outliers_series.ww.init()

    has_outliers_dict = outliers_series.ww.box_plot_dict()
    assert has_outliers_dict['low_bound'] == 8.125
    assert has_outliers_dict['high_bound'] == 83.125
    assert has_outliers_dict['quantiles'] == {0.0: -16.0, 0.25: 36.25, 0.5: 42.0, 0.75: 55.0, 1.0: 93.0}
    assert has_outliers_dict['low_values'] == [-16]
    assert has_outliers_dict['high_values'] == [93]
    assert has_outliers_dict['low_indices'] == [3]
    assert has_outliers_dict['high_indices'] == [0]

    no_outliers_dict = no_outliers_series.ww.box_plot_dict()

    assert no_outliers_dict['low_bound'] == 8.125
    assert no_outliers_dict['high_bound'] == 83.125
    assert no_outliers_dict['quantiles'] == {0.0: 23.0, 0.25: 36.25, 0.5: 42.0, 0.75: 55.0, 1.0: 60.0}
    assert no_outliers_dict['low_values'] == []
    assert no_outliers_dict['high_values'] == []
    assert no_outliers_dict['low_indices'] == []
    assert no_outliers_dict['high_indices'] == []


def test_box_plot_outliers_with_quantiles(outliers_df):
    outliers_series = outliers_df['has_outliers']
    outliers_series.ww.init()

    no_outliers_series = outliers_df['no_outliers']
    no_outliers_series.ww.init()

    has_outliers_dict = outliers_series.ww.box_plot_dict(quantiles={0.0: -16.0, 0.25: 36.25, 0.5: 42.0, 0.75: 55.0, 1.0: 93.0})
    assert has_outliers_dict['low_bound'] == 8.125
    assert has_outliers_dict['high_bound'] == 83.125
    assert has_outliers_dict['quantiles'] == {0.0: -16.0, 0.25: 36.25, 0.5: 42.0, 0.75: 55.0, 1.0: 93.0}
    assert has_outliers_dict['low_values'] == [-16]
    assert has_outliers_dict['high_values'] == [93]
    assert has_outliers_dict['low_indices'] == [3]
    assert has_outliers_dict['high_indices'] == [0]

    no_outliers_dict = no_outliers_series.ww.box_plot_dict(quantiles={0.0: 23.0, 0.25: 36.25, 0.5: 42.0, 0.75: 55.0, 1.0: 60.0})

    assert no_outliers_dict['low_bound'] == 8.125
    assert no_outliers_dict['high_bound'] == 83.125
    assert no_outliers_dict['quantiles'] == {0.0: 23.0, 0.25: 36.25, 0.5: 42.0, 0.75: 55.0, 1.0: 60.0}
    assert no_outliers_dict['low_values'] == []
    assert no_outliers_dict['high_values'] == []
    assert no_outliers_dict['low_indices'] == []
    assert no_outliers_dict['high_indices'] == []


def test_get_outliers_for_column_with_nans(outliers_df):
    contains_nans_series = outliers_df['has_outliers_with_nans']
    contains_nans_series.ww.init()

    box_plot_dict = contains_nans_series.ww.box_plot_dict()

    assert box_plot_dict['low_bound'] == 4.5
    assert box_plot_dict['high_bound'] == 88.5
    assert box_plot_dict['quantiles'] == {0.0: -16.0, 0.25: 36.0, 0.5: 42.0, 0.75: 57.0, 1.0: 93.0}
    assert box_plot_dict['low_values'] == [-16]
    assert box_plot_dict['high_values'] == [93]
    assert box_plot_dict['low_indices'] == [3]
    assert box_plot_dict['high_indices'] == [5]


def test_box_plot_on_non_numeric_col(outliers_df):
    error = "Cannot calculate box plot statistics for non-numeric column"

    non_numeric_series = init_series(outliers_df['non_numeric'], logical_type='Categorical')
    with pytest.raises(TypeError, match=error):
        non_numeric_series.ww.box_plot_dict()

    wrong_dtype_series = init_series(outliers_df['has_outliers'], logical_type='Categorical')
    with pytest.raises(TypeError, match=error):
        wrong_dtype_series.ww.box_plot_dict()


def test_box_plot_with_fully_null_col(outliers_df):
    fully_null_double_series = init_series(outliers_df['nans'], logical_type='Double')

    box_plot_dict = fully_null_double_series.ww.box_plot_dict()
    check_empty_box_plot_dict(box_plot_dict)

    box_plot_dict = fully_null_double_series.ww.box_plot_dict(quantiles={0.25: 1, 0.75: 10})
    check_empty_box_plot_dict(box_plot_dict)

    fully_null_int_series = init_series(outliers_df['nans'], logical_type='IntegerNullable')
    box_plot_dict = fully_null_int_series.ww.box_plot_dict()
    check_empty_box_plot_dict(box_plot_dict)

    fully_null_categorical_series = init_series(outliers_df['nans'], logical_type='Categorical')
    error = "Cannot calculate box plot statistics for non-numeric column"
    with pytest.raises(TypeError, match=error):
        fully_null_categorical_series.ww.box_plot_dict()


def test_box_plot_with_empty_col(outliers_df):
    series = outliers_df['nans'].dropna()

    fully_null_double_series = init_series(series, logical_type='Double')

    box_plot_dict = fully_null_double_series.ww.box_plot_dict()
    check_empty_box_plot_dict(box_plot_dict)

    box_plot_dict = fully_null_double_series.ww.box_plot_dict(quantiles={0.25: 1, 0.75: 10})
    check_empty_box_plot_dict(box_plot_dict)

    fully_null_int_series = init_series(series, logical_type='IntegerNullable')
    box_plot_dict = fully_null_int_series.ww.box_plot_dict()
    check_empty_box_plot_dict(box_plot_dict)

    fully_null_categorical_series = init_series(series, logical_type='Categorical')
    error = "Cannot calculate box plot statistics for non-numeric column"
    with pytest.raises(TypeError, match=error):
        fully_null_categorical_series.ww.box_plot_dict()


def test_box_plot_different_quantiles(outliers_df):
    has_outliers_series = outliers_df['has_outliers']
    has_outliers_series.ww.init()

    box_plot_info = has_outliers_series.ww.box_plot_dict()

    assert set(box_plot_info.keys()) == {'low_bound',
                                         'high_bound',
                                         'quantiles',
                                         'low_values',
                                         'high_values',
                                         'low_indices',
                                         'high_indices'}
    assert box_plot_info['low_bound'] == 8.125
    assert box_plot_info['high_bound'] == 83.125
    assert len(box_plot_info['quantiles']) == 5
    assert len(box_plot_info['high_values']) == 1
    assert len(box_plot_info['low_values']) == 1

    box_plot_info = has_outliers_series.ww.box_plot_dict(quantiles={0.25: 36.25, 0.75: 55.0})

    assert set(box_plot_info.keys()) == {'low_bound',
                                         'high_bound',
                                         'quantiles',
                                         'low_values',
                                         'high_values',
                                         'low_indices',
                                         'high_indices'}
    assert box_plot_info['low_bound'] == 8.125
    assert box_plot_info['high_bound'] == 83.125
    assert len(box_plot_info['quantiles']) == 2
    assert len(box_plot_info['high_values']) == 1
    assert len(box_plot_info['low_values']) == 1

    partial_quantiles = {0.0: -16, 0.25: 36.25, 0.75: 55.0, 1.0: 93}
    box_plot_info = has_outliers_series.ww.box_plot_dict(quantiles=partial_quantiles)

    assert set(box_plot_info.keys()) == {'low_bound',
                                         'high_bound',
                                         'quantiles',
                                         'low_values',
                                         'high_values',
                                         'low_indices',
                                         'high_indices'}
    assert box_plot_info['low_bound'] == 8.125
    assert box_plot_info['high_bound'] == 83.125
    assert len(box_plot_info['quantiles']) == 4
    assert len(box_plot_info['high_values']) == 1
    assert len(box_plot_info['low_values']) == 1


def test_box_plot_quantiles_errors(outliers_df):
    series = outliers_df['has_outliers']
    series.ww.init()

    error = re.escape("Input quantiles do not contain the minimum necessary quantiles for outlier calculation: "
                      "0.25 (the first quartile) and 0.75 (the third quartile).")

    partial_quantiles = {0.25: 36.25}
    with pytest.raises(ValueError, match=error):
        series.ww.box_plot_dict(quantiles=partial_quantiles)

    empty_quantiles = {}
    with pytest.raises(ValueError, match=error):
        series.ww.box_plot_dict(quantiles=empty_quantiles)

    error = "quantiles must be a dictionary."
    with pytest.raises(TypeError, match=error):
        series.ww.box_plot_dict(quantiles=1)




import re

import numpy as np
import pandas as pd
import pytest
from mock import patch

import woodwork as ww
from woodwork.logical_types import (
    Age,
    AgeNullable,
    Boolean,
    BooleanNullable,
    Categorical,
    CountryCode,
    Datetime,
    Double,
    Integer,
    IntegerNullable,
    Ordinal,
    PostalCode,
    SubRegionCode
)
from woodwork.type_sys.type_system import DEFAULT_INFERENCE_FUNCTIONS
from woodwork.type_sys.utils import (
    _get_specified_ltype_params,
    _is_categorical_series,
    _is_numeric_series,
    col_is_datetime,
    list_logical_types,
    list_semantic_tags
)
from woodwork.utils import (
    _convert_input_to_set,
    _get_column_logical_type,
    _infer_datetime_format,
    _is_null_latlong,
    _is_s3,
    _is_url,
    _is_valid_latlong_series,
    _is_valid_latlong_value,
    _parse_logical_type,
    _reformat_to_latlong,
    _to_latlong_float,
    camel_to_snake,
    get_valid_mi_types,
    import_or_none,
    import_or_raise
)

dd = import_or_none('dask.dataframe')
ks = import_or_none('databricks.koalas')


def test_camel_to_snake():
    test_items = {
        'PostalCode': 'postal_code',
        'SubRegionCode': 'sub_region_code',
        'NaturalLanguage': 'natural_language',
        'Categorical': 'categorical',
    }

    for key, value in test_items.items():
        assert camel_to_snake(key) == value


def test_convert_input_to_set():
    error_message = "semantic_tags must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        _convert_input_to_set(int)

    error_message = "test_text must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        _convert_input_to_set({'index': {}, 'time_index': {}}, 'test_text')

    error_message = "include parameter must contain only strings"
    with pytest.raises(TypeError, match=error_message):
        _convert_input_to_set(['index', 1], 'include parameter')

    semantic_tags_from_single = _convert_input_to_set('index', 'include parameter')
    assert semantic_tags_from_single == {'index'}

    semantic_tags_from_list = _convert_input_to_set(['index', 'numeric', 'category'])
    assert semantic_tags_from_list == {'index', 'numeric', 'category'}

    semantic_tags_from_set = _convert_input_to_set({'index', 'numeric', 'category'}, 'include parameter')
    assert semantic_tags_from_set == {'index', 'numeric', 'category'}


@patch("woodwork.utils._validate_string_tags")
@patch("woodwork.utils._validate_tags_input_type")
def test_validation_methods_called(mock_validate_input_type, mock_validate_strings):
    assert not mock_validate_input_type.called
    assert not mock_validate_strings.called

    _convert_input_to_set('test_tag', validate=False)
    assert not mock_validate_input_type.called

    _convert_input_to_set('test_tag', validate=True)
    assert mock_validate_input_type.called

    _convert_input_to_set(['test_tag', 'tag2'], validate=False)
    assert not mock_validate_strings.called

    _convert_input_to_set(['test_tag', 'tag2'], validate=True)
    assert mock_validate_strings.called


def test_list_logical_types_default():
    all_ltypes = ww.logical_types.LogicalType.__subclasses__()

    df = list_logical_types()

    assert set(df.columns) == {'name', 'type_string', 'description', 'physical_type',
                               'standard_tags', 'is_default_type', 'is_registered', 'parent_type'}

    assert len(all_ltypes) == len(df)
    default_types_set = {str(cls) for cls in DEFAULT_INFERENCE_FUNCTIONS.keys()}
    listed_as_default = set(df[df['is_default_type']]['name'])
    assert listed_as_default == default_types_set


def test_list_logical_types_customized_type_system():
    ww.type_system.remove_type('URL')

    class CustomRegistered(ww.logical_types.LogicalType):
        primary_dtype = 'int64'

    class CustomNotRegistered(ww.logical_types.LogicalType):
        primary_dtype = 'int64'

    ww.type_system.add_type(CustomRegistered)
    all_ltypes = ww.logical_types.LogicalType.__subclasses__()
    df = list_logical_types()
    assert len(all_ltypes) == len(df)
    # Check that URL is unregistered
    url = df[df.name == 'URL'].iloc[0]
    assert url.is_default_type
    assert not url.is_registered

    # Check that new registered type is present and shows as registered
    index = df.name == 'CustomRegistered'
    assert index.any()
    custom = df[index].iloc[0]
    assert not custom.is_default_type
    assert custom.is_registered

    # Check that new unregistered type is present and shows as not registered
    index = df.name == 'CustomNotRegistered'
    assert index.any()
    custom = df[index].iloc[0]
    assert not custom.is_default_type
    assert not custom.is_registered
    ww.type_system.reset_defaults()


def test_list_semantic_tags():
    df = list_semantic_tags()

    assert set(df.columns) == {'name', 'is_standard_tag', 'valid_logical_types'}

    for name, log_type_list in df[['name', 'valid_logical_types']].values:
        if name not in ['index', 'time_index', 'date_of_birth']:
            for log_type in log_type_list:
                assert name in log_type.standard_tags


def test_is_numeric_datetime_series(time_index_df):
    assert _is_numeric_series(time_index_df['ints'], None)
    assert _is_numeric_series(time_index_df['ints'], Double)
    assert not _is_numeric_series(time_index_df['ints'], Categorical)
    assert _is_numeric_series(time_index_df['ints'], Datetime)

    assert not _is_numeric_series(time_index_df['strs'], None)
    assert not _is_numeric_series(time_index_df['strs'], 'Categorical')
    assert not _is_numeric_series(time_index_df['strs'], Categorical)
    assert _is_numeric_series(time_index_df['strs'], Double)
    assert _is_numeric_series(time_index_df['strs'], 'Double')

    assert not _is_numeric_series(time_index_df['bools'], None)
    assert not _is_numeric_series(time_index_df['bools'], 'Boolean')

    assert not _is_numeric_series(time_index_df['times'], None)
    assert not _is_numeric_series(time_index_df['times'], Datetime)

    assert not _is_numeric_series(time_index_df['letters'], None)
    assert not _is_numeric_series(time_index_df['letters'], Double)
    assert not _is_numeric_series(time_index_df['letters'], Categorical)


def test_get_ltype_params():
    params_empty_class = _get_specified_ltype_params(Categorical)
    assert params_empty_class == {}
    params_empty = _get_specified_ltype_params(Categorical())
    assert params_empty == {}

    params_class = _get_specified_ltype_params(Datetime)
    assert params_class == {}

    params_null = _get_specified_ltype_params(Datetime())
    assert params_null == {'datetime_format': None}

    ymd = '%Y-%m-%d'
    params_value = _get_specified_ltype_params(Datetime(datetime_format=ymd))
    assert params_value == {'datetime_format': ymd}


def test_import_or_raise():
    assert import_or_raise('pandas', 'Module pandas could not be found') == pd

    error = 'Module nonexistent could not be found.'
    with pytest.raises(ImportError, match=error):
        import_or_raise('nonexistent', error)


def test_import_or_none():
    assert import_or_none('pandas') == pd
    assert import_or_none('nonexistent') is None


def test_is_url():
    assert _is_url('https://www.google.com/')
    assert not _is_url('google.com')


def test_is_s3():
    assert _is_s3('s3://test-bucket/test-key')
    assert not _is_s3('https://woodwork-static.s3.amazonaws.com/')


def test_reformat_to_latlong_errors():
    for latlong in [{1, 2, 3}, '{1, 2, 3}', 'This is text']:
        error = (f'LatLongs must either be a tuple, a list, or a string representation of a tuple. {latlong} does not fit the criteria.')
        with pytest.raises(ValueError, match=error):
            _reformat_to_latlong(latlong)

    error = re.escape("LatLongs must either be a tuple, a list, or a string representation of a tuple. (1,2) does not fit the criteria.")
    with pytest.raises(ValueError, match=error):
        _reformat_to_latlong("'(1,2)'")

    for latlong in [(1, 2, 3), '(1, 2, 3)']:
        error = re.escape("LatLong values must have exactly two values. (1, 2, 3) does not have two values.")
        with pytest.raises(ValueError, match=error):
            _reformat_to_latlong(latlong)

    error = re.escape("Latitude and Longitude values must be in decimal degrees. The latitude or longitude represented by 41deg52\'54\" N cannot be converted to a float.")
    with pytest.raises(ValueError, match=error):
        _reformat_to_latlong(('41deg52\'54\" N', '21deg22\'54\" W'))


def test_reformat_to_latlong():
    simple_latlong = (1, 2)

    assert _reformat_to_latlong((1, 2)) == simple_latlong
    assert _reformat_to_latlong(('1', '2')) == simple_latlong
    assert _reformat_to_latlong('(1,2)') == simple_latlong

    # Check non-standard tuple formats
    assert _reformat_to_latlong([1, 2]) == simple_latlong
    assert _reformat_to_latlong(['1', '2']) == simple_latlong
    assert _reformat_to_latlong('[1, 2]') == simple_latlong
    assert _reformat_to_latlong('1, 2') == simple_latlong

    assert _reformat_to_latlong(None) is np.nan
    assert _reformat_to_latlong((1, np.nan)) == (1, np.nan)
    assert _reformat_to_latlong((np.nan, '1')) == (np.nan, 1)

    # This is how csv and parquet will deserialize
    assert _reformat_to_latlong('(1, nan)') == (1, np.nan)
    assert _reformat_to_latlong('(NaN, 9)') == (np.nan, 9)


def test_reformat_to_latlong_list():
    simple_latlong = [1, 2]

    assert _reformat_to_latlong((1, 2), use_list=True) == simple_latlong
    assert _reformat_to_latlong(('1', '2'), use_list=True) == simple_latlong
    assert _reformat_to_latlong('(1,2)', use_list=True) == simple_latlong

    assert _reformat_to_latlong([1, 2], use_list=True) == simple_latlong
    assert _reformat_to_latlong(['1', '2'], use_list=True) == simple_latlong
    assert _reformat_to_latlong('[1, 2]', use_list=True) == simple_latlong
    assert _reformat_to_latlong('1, 2', use_list=True) == simple_latlong

    assert _reformat_to_latlong((1, np.nan), use_list=True) == [1, np.nan]
    assert _reformat_to_latlong((np.nan, '1'), use_list=True) == [np.nan, 1]

    # This is how csv and parquet will deserialize
    assert _reformat_to_latlong('[1, nan]', use_list=True) == [1, np.nan]
    assert _reformat_to_latlong('[1, NaN]', use_list=True) == [1, np.nan]


def test_to_latlong_float():
    assert _to_latlong_float(4) == 4.0
    assert _to_latlong_float('2.2') == 2.2

    assert _to_latlong_float(None) is np.nan
    assert _to_latlong_float(np.nan) is np.nan
    assert _to_latlong_float(pd.NA) is np.nan

    error = re.escape('Latitude and Longitude values must be in decimal degrees. The latitude or longitude represented by [1, 2, 3] cannot be converted to a float.')
    with pytest.raises(ValueError, match=error):
        _to_latlong_float([1, 2, 3])


def test_is_null_latlong():
    assert _is_null_latlong(None)
    assert _is_null_latlong(np.nan)
    assert _is_null_latlong(pd.NA)
    assert _is_null_latlong('None')
    assert _is_null_latlong('nan')
    assert _is_null_latlong('NaN')

    assert not _is_null_latlong([None, 1, 3])
    assert not _is_null_latlong('none')
    assert not _is_null_latlong(0)
    assert not _is_null_latlong(False)


def test_is_valid_latlong_value():
    values = [
        (1.0, 2.0),
        np.nan,
        [1.0, 2.0],
        (np.nan, np.nan),
        ('a', 2.0),
        (1.0, 2.0, 3.0),
        None
    ]

    expected_values = [
        True,
        True,
        False,
        False,
        False,
        False,
        False
    ]

    for index, value in enumerate(values):
        assert _is_valid_latlong_value(value) is expected_values[index]


def test_is_valid_latlong_value_koalas():
    values = [
        (1.0, 2.0),
        np.nan,
        [1.0, 2.0],
        (np.nan, np.nan),
        ('a', 2.0),
        (1.0, 2.0, 3.0),
        None
    ]

    expected_values = [
        False,
        True,
        True,
        False,
        False,
        False,
        False
    ]

    for index, value in enumerate(values):
        assert _is_valid_latlong_value(value, bracket_type=list) is expected_values[index]


def test_is_valid_latlong_series():
    valid_series = pd.Series([(1.0, 2.0), (3.0, 4.0)])
    invalid_series = pd.Series([(1.0, 2.0), (3.0, '4.0')])

    assert _is_valid_latlong_series(valid_series) is True
    assert _is_valid_latlong_series(invalid_series) is False


def test_get_valid_mi_types():
    valid_types = get_valid_mi_types()
    expected_types = [
        Age,
        AgeNullable,
        Boolean,
        BooleanNullable,
        Categorical,
        CountryCode,
        Datetime,
        Double,
        Integer,
        IntegerNullable,
        Ordinal,
        PostalCode,
        SubRegionCode,
    ]

    assert valid_types == expected_types


def test_get_column_logical_type(sample_series):
    assert isinstance(_get_column_logical_type(sample_series, None, 'col_name'), Categorical)

    assert isinstance(_get_column_logical_type(sample_series, Datetime, 'col_name'), Datetime)


def test_parse_logical_type():
    assert isinstance(_parse_logical_type('Datetime', 'col_name'), Datetime)
    assert isinstance(_parse_logical_type(Datetime, 'col_name'), Datetime)

    ymd_format = Datetime(datetime_format='%Y-%m-%d')
    assert _parse_logical_type(ymd_format, 'col_name') == ymd_format


def test_parse_logical_type_errors():
    error = "Invalid logical type specified for 'col_name'"
    with pytest.raises(TypeError, match=error):
        _parse_logical_type(int, 'col_name')


def test_col_is_datetime():
    inputs = [
        pd.to_datetime(pd.Series(['2020-01-01', '2021-02-02', '2022-03-03'])),
        pd.to_datetime(pd.Series([pd.NA, '2021-02-02', '2022-03-03'])),
        pd.Series([1, 2, 3]),
        pd.Series([pd.NA, 2, 3]),
        pd.Series([1.0, 2.0, 3.0]),
        pd.Series([pd.NA, 2.0, 3.0]),
        pd.Series(['2020-01-01', '2021-02-02', '2022-03-03']),
        pd.Series([pd.NA, '2021-02-02', '2022-03-03']),
        pd.Series(['a', 'b', 'c']),
        pd.Series([pd.NA, 'b', 'c']),
        pd.Series([pd.NA, pd.NA, pd.NA]),
    ]

    expected_values = [
        True,
        True,
        False,
        False,
        False,
        False,
        True,
        True,
        False,
        False,
        False,
    ]

    for input, expected in list(zip(inputs, expected_values)):
        actual = col_is_datetime(input)
        assert actual is expected


def test_infer_datetime_format(datetimes):
    for series in datetimes:
        fmt = _infer_datetime_format(series)
        assert fmt == '%m/%d/%Y'

    dt1 = pd.Series(['3/11/2000 9:00', '3/11/2000 10:00', '3/11/2000 11:00', '3/11/2000 12:00'])
    fmt = _infer_datetime_format(dt1)
    assert fmt == '%m/%d/%Y %H:%M'


def test_infer_datetime_format_all_null():
    missing_data = [
        pd.Series([None, None, None]),
        pd.Series([np.nan, np.nan, np.nan]),
        pd.Series([pd.NA, pd.NA, pd.NA]),
        pd.Series([]),
    ]

    for pd_series in missing_data:
        assert _infer_datetime_format(pd_series) is None
        if dd:
            dd_series = dd.from_pandas(pd_series, npartitions=2)
            assert _infer_datetime_format(dd_series) is None
        if ks:
            ks_series = ks.from_pandas(pd_series)
            assert _infer_datetime_format(ks_series) is None


def test_is_categorical() -> None:
    # not categorical because unhashable type (list)
    assert not _is_categorical_series(pd.Series([[1]]), 0)
    assert not _is_categorical_series(pd.Series([None, [1]]), 0)

    # not categorical because empty series
    assert not _is_categorical_series(pd.Series([]), 0)
    assert not _is_categorical_series(pd.Series([None]), 0)
    assert not _is_categorical_series(pd.Series([None, None]), 0)

    # not categorical because too many unique values
    assert not _is_categorical_series(pd.Series([1, 2]), 0.5)
    assert not _is_categorical_series(pd.Series([1, 2, 3, 1]), 0.5)
    assert not _is_categorical_series(pd.Series([1, 2, 3, 4]), 0.75)

    # categorical
    assert _is_categorical_series(pd.Series([1, 1]), 0.5)
    assert _is_categorical_series(pd.Series([1, 2, 1, 1]), 0.5)
    assert _is_categorical_series(pd.Series([1, 2, 3, 1]), 0.75)




import pandas as pd

import woodwork as ww
from woodwork.logical_types import (
    Boolean,
    Categorical,
    Datetime,
    Double,
    Integer,
    NaturalLanguage
)


def load_retail(id='demo_retail_data', nrows=None, init_woodwork=True):
    """Load a demo retail dataset into a DataFrame, optionally initializing Woodwork's typing information.

    Args:
        id (str, optional): The name to assign to the DataFrame, if returning a DataFrame with Woodwork
            typing information initialized. If not returning a DataFrame with Woodwork initialized,
            this will be ignored. Defaults to ``demo_retail_data``.
        nrows (int, optional): The number of rows to return in the dataset. If None, will
            return all possible rows. Defaults to None.
        init_woodwork (bool): If True, will return a pandas DataFrame with Woodwork
            typing information initialized. If False, will return a DataFrame without
            Woodwork initialized. Defaults to False.

    Returns:
        pd.DataFrame: A DataFrame containing the demo data with Woodwork typing initialized.
        If `init_woodwork` is False, will return an uninitialized DataFrame.
    """
    csv_s3_gz = "https://api.featurelabs.com/datasets/online-retail-logs-2018-08-28.csv.gz?library=woodwork&version=" + ww.__version__
    csv_s3 = "https://api.featurelabs.com/datasets/online-retail-logs-2018-08-28.csv?library=woodwork&version=" + ww.__version__
    # Try to read in gz compressed file
    try:
        df = pd.read_csv(csv_s3_gz,
                         nrows=nrows,
                         parse_dates=["order_date"])
    # Fall back to uncompressed
    except Exception:
        df = pd.read_csv(csv_s3,
                         nrows=nrows,
                         parse_dates=["order_date"])
    # Add unique column for index
    df.insert(0, 'order_product_id', range(len(df)))

    if init_woodwork:
        logical_types = {
            'order_product_id': Categorical,
            'order_id': Categorical,
            'product_id': Categorical,
            'description': NaturalLanguage,
            'quantity': Integer,
            'order_date': Datetime,
            'unit_price': Double,
            'customer_name': Categorical,
            'country': Categorical,
            'total': Double,
            'cancelled': Boolean,
        }

        df.ww.init(name=id,
                   index='order_product_id',
                   time_index='order_date',
                   logical_types=logical_types)

    return df



import contextlib

CONFIG_DEFAULTS = {
    'categorical_threshold': 0.2,
    'numeric_categorical_threshold': None,
    'email_inference_regex': r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)",
}


class Config:
    def __init__(self, default_values):
        self._defaults = default_values
        self._data = default_values.copy()

    def set_option(self, key, value):
        if key not in self._data.keys():
            raise KeyError(f"Invalid option specified: {key}")
        self._data[key] = value

    def get_option(self, key):
        if key not in self._data.keys():
            raise KeyError(f"Invalid option specified: {key}")
        return self._data[key]

    def reset_option(self, key):
        if key not in self._data.keys():
            raise KeyError(f"Invalid option specified: {key}")
        self._data[key] = self._defaults[key]

    @contextlib.contextmanager
    def with_options(self, **options):
        old_options = {k: self.get_option(k) for k in options}

        for k, v in options.items():
            self.set_option(k, v)
        try:
            yield
        finally:
            for k, v in old_options.items():
                self.set_option(k, v)

    def __repr__(self):
        output_string = "Woodwork Global Config Settings\n"
        output_string += "-" * (len(output_string) - 1)
        for key, value in self._data.items():
            output_string += f"\n{key}: {value}"
        return output_string


config = Config(CONFIG_DEFAULTS)



import re

import numpy as np
import pandas as pd
import pytest
from mock import patch

from woodwork.accessor_utils import (
    _is_dask_series,
    _is_dataframe,
    _is_koalas_series,
    init_series
)
from woodwork.column_accessor import WoodworkColumnAccessor
from woodwork.column_schema import ColumnSchema
from woodwork.exceptions import (
    ParametersIgnoredWarning,
    TypeConversionError,
    TypingInfoMismatchWarning,
    WoodworkNotInitError
)
from woodwork.logical_types import (
    Categorical,
    CountryCode,
    Double,
    Integer,
    LatLong,
    NaturalLanguage,
    Ordinal,
    PostalCode,
    SubRegionCode
)
from woodwork.tests.testing_utils import (
    is_property,
    is_public_method,
    to_pandas
)
from woodwork.utils import import_or_none

dd = import_or_none('dask.dataframe')
ks = import_or_none('databricks.koalas')


def test_accessor_init(sample_series):
    assert sample_series.ww._schema is None
    sample_series.ww.init()
    assert isinstance(sample_series.ww._schema, ColumnSchema)
    assert isinstance(sample_series.ww.logical_type, Categorical)
    assert sample_series.ww.semantic_tags == {'category'}


def test_accessor_init_with_schema(sample_series):
    sample_series.ww.init(semantic_tags={'test_tag'},
                          description='this is a column',
                          origin='base')
    schema = sample_series.ww.schema

    head_series = sample_series.head(2)
    assert head_series.ww.schema is None
    head_series.ww.init(schema=schema)

    assert head_series.ww._schema is schema
    assert head_series.ww.description == 'this is a column'
    assert head_series.ww.origin == 'base'
    assert head_series.ww.semantic_tags == {'test_tag', 'category'}

    iloc_series = sample_series.loc[[2, 3]]
    assert iloc_series.ww.schema is None
    iloc_series.ww.init(schema=schema)

    assert iloc_series.ww._schema is schema
    assert iloc_series.ww.description == 'this is a column'
    assert iloc_series.ww.origin == 'base'
    assert iloc_series.ww.semantic_tags == {'test_tag', 'category'}
    assert isinstance(iloc_series.ww.logical_type, Categorical)


def test_accessor_init_with_schema_errors(sample_series):
    sample_series.ww.init(semantic_tags={'test_tag'},
                          description='this is a column',
                          origin='base')
    schema = sample_series.ww.schema

    head_series = sample_series.head(2)
    assert head_series.ww.schema is None

    error = 'Provided schema must be a Woodwork.ColumnSchema object.'
    with pytest.raises(TypeError, match=error):
        head_series.ww.init(schema=int)

    if _is_koalas_series(sample_series):
        ltype_dtype = 'string'
        new_dtype = '<U0'
    else:
        ltype_dtype = 'category'
        new_dtype = 'object'

    error = re.escape(f"dtype mismatch between Series dtype {new_dtype}, and Categorical dtype, {ltype_dtype}")
    diff_dtype_series = sample_series.astype(new_dtype)
    with pytest.raises(ValueError, match=error):
        diff_dtype_series.ww.init(schema=schema)


def test_accessor_with_schema_parameter_warning(sample_series):
    sample_series.ww.init(semantic_tags={'test_tag'}, description='this is a column', origin='base')
    schema = sample_series.ww.schema

    head_series = sample_series.head(2)

    warning = "A schema was provided and the following parameters were ignored: " \
              "logical_type, semantic_tags, description, origin, metadata, use_standard_tags"
    with pytest.warns(ParametersIgnoredWarning, match=warning):
        head_series.ww.init(logical_type=Integer, semantic_tags={'ignored_test_tag'},
                            description='an ignored description', origin='ignored origin',
                            metadata={'user_id': 'user1'}, use_standard_tags=False, schema=schema)

    assert head_series.ww.semantic_tags == {'category', 'test_tag'}
    assert isinstance(head_series.ww.logical_type, Categorical)
    assert head_series.ww.description == 'this is a column'
    assert head_series.ww.origin == 'base'
    assert head_series.ww.metadata == {}


def test_accessor_init_with_logical_type(sample_series):
    series = sample_series.astype('string')
    series.ww.init(logical_type=NaturalLanguage)
    assert isinstance(series.ww.logical_type, NaturalLanguage)
    assert series.ww.semantic_tags == set()

    series = sample_series.astype('string')
    series.ww.init(logical_type="natural_language")
    assert isinstance(series.ww.logical_type, NaturalLanguage)
    assert series.ww.semantic_tags == set()

    series = sample_series.astype('string')
    series.ww.init(logical_type="NaturalLanguage")
    assert isinstance(series.ww.logical_type, NaturalLanguage)
    assert series.ww.semantic_tags == set()


def test_accessor_init_with_invalid_logical_type(sample_series):
    if _is_koalas_series(sample_series):
        series_dtype = '<U0'
    else:
        series_dtype = 'object'
    series = sample_series.astype(series_dtype)
    correct_dtype = 'string'
    error_message = re.escape(f"Cannot initialize Woodwork. Series dtype '{series_dtype}' is incompatible with "
                              f"NaturalLanguage dtype. Try converting series dtype to '{correct_dtype}' before "
                              "initializing or use the woodwork.init_series function to initialize.")
    with pytest.raises(ValueError, match=error_message):
        series.ww.init(logical_type=NaturalLanguage)


def test_accessor_init_with_semantic_tags(sample_series):
    semantic_tags = ['tag1', 'tag2']
    sample_series.ww.init(semantic_tags=semantic_tags, use_standard_tags=False)
    assert sample_series.ww.semantic_tags == set(semantic_tags)


def test_error_accessing_properties_before_init(sample_series):
    props_to_exclude = ['iloc', 'loc', 'schema', '_series']
    props = [prop for prop in dir(sample_series.ww) if is_property(WoodworkColumnAccessor, prop) and prop not in props_to_exclude]

    error = "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    for prop in props:
        with pytest.raises(WoodworkNotInitError, match=error):
            getattr(sample_series.ww, prop)


def test_error_accessing_methods_before_init(sample_series):
    methods_to_exclude = ['init']
    public_methods = [method for method in dir(sample_series.ww) if is_public_method(WoodworkColumnAccessor, method)]
    public_methods = [method for method in public_methods if method not in methods_to_exclude]

    method_args_dict = {
        'add_semantic_tags': [{'new_tag'}],
        'remove_semantic_tags': [{'new_tag'}],
        'reset_semantic_tags': None,
        'set_logical_type': ['Integer'],
        'set_semantic_tags': [{'new_tag'}],
        'box_plot_dict': [None],
    }
    error = "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    for method in public_methods:
        func = getattr(sample_series.ww, method)
        method_args = method_args_dict[method]
        with pytest.raises(WoodworkNotInitError, match=error):
            if method_args:
                func(*method_args)
            else:
                func()


def test_accessor_with_alternate_semantic_tags_input(sample_series):
    series = sample_series.copy()
    semantic_tags = 'custom_tag'
    series.ww.init(semantic_tags=semantic_tags, use_standard_tags=False)
    assert series.ww.semantic_tags == {'custom_tag'}

    series = sample_series.copy()
    semantic_tags = {'custom_tag', 'numeric'}
    series.ww.init(semantic_tags=semantic_tags, use_standard_tags=False)
    assert series.ww.semantic_tags == semantic_tags


def test_logical_type_errors(sample_series):
    error_message = "Invalid logical type specified for 'sample_series'"
    with pytest.raises(TypeError, match=error_message):
        sample_series.ww.init(logical_type=int)

    error_message = "String naturalllanguage is not a valid logical type"
    with pytest.raises(ValueError, match=error_message):
        sample_series.ww.init(logical_type='naturalllanguage')


def test_semantic_tag_errors(sample_series):
    error_message = "semantic_tags must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        sample_series.ww.init(semantic_tags=int)

    error_message = "semantic_tags must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        sample_series.ww.init(semantic_tags={'index': {}, 'time_index': {}})

    error_message = "semantic_tags must contain only strings"
    with pytest.raises(TypeError, match=error_message):
        sample_series.ww.init(semantic_tags=['index', 1])


def test_accessor_description(sample_series):
    column_description = "custom description"
    sample_series.ww.init(description=column_description)
    assert sample_series.ww.description == column_description

    new_description = "updated description text"
    sample_series.ww.description = new_description
    assert sample_series.ww.description == new_description


def test_description_setter_error_before_init(sample_series):
    err_msg = "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    with pytest.raises(WoodworkNotInitError, match=err_msg):
        sample_series.ww.description = "description"


def test_description_error_on_init(sample_series):
    err_msg = "Column description must be a string"
    with pytest.raises(TypeError, match=err_msg):
        sample_series.ww.init(description=123)


def test_description_error_on_update(sample_series):
    sample_series.ww.init()
    err_msg = "Column description must be a string"
    with pytest.raises(TypeError, match=err_msg):
        sample_series.ww.description = 123


def test_accessor_origin(sample_series):
    column_origin = "base"
    sample_series.ww.init(origin=column_origin)
    assert sample_series.ww.origin == column_origin

    new_origin = "engineered"
    sample_series.ww.origin = new_origin
    assert sample_series.ww.origin == new_origin


def test_origin_setter_error_before_init(sample_series):
    err_msg = "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    with pytest.raises(WoodworkNotInitError, match=err_msg):
        sample_series.ww.origin = "base"


def test_origin_error_on_init(sample_series):
    err_msg = "Column origin must be a string"
    with pytest.raises(TypeError, match=err_msg):
        sample_series.ww.init(origin=123)


def test_origin_error_on_update(sample_series):
    sample_series.ww.init()
    err_msg = "Column origin must be a string"
    with pytest.raises(TypeError, match=err_msg):
        sample_series.ww.origin = 123


def test_accessor_repr(sample_series):
    sample_series.ww.init(use_standard_tags=False)
    # Koalas doesn't support categorical
    if _is_koalas_series(sample_series):
        dtype = 'string'
    else:
        dtype = 'category'
    assert sample_series.ww.__repr__() == f'<Series: sample_series (Physical Type = {dtype}) ' \
        '(Logical Type = Categorical) (Semantic Tags = set())>'


def test_accessor_repr_error_before_init(sample_series):
    err_msg = "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    with pytest.raises(WoodworkNotInitError, match=err_msg):
        sample_series.ww.__repr__()


def test_set_semantic_tags(sample_series):
    semantic_tags = {'tag1', 'tag2'}
    sample_series.ww.init(semantic_tags=semantic_tags)
    assert sample_series.ww.semantic_tags == semantic_tags.union({'category'})

    new_tags = ['new_tag']
    sample_series.ww.set_semantic_tags(new_tags)
    assert sample_series.ww.semantic_tags == set(new_tags).union({'category'})


def test_adds_numeric_standard_tag():
    series = pd.Series([1, 2, 3])
    semantic_tags = 'custom_tag'

    logical_types = [Integer, Double]
    for logical_type in logical_types:
        series = series.astype(logical_type.primary_dtype)
        series.ww.init(logical_type=logical_type, semantic_tags=semantic_tags)
        assert series.ww.semantic_tags == {'custom_tag', 'numeric'}


def test_adds_category_standard_tag():
    semantic_tags = 'custom_tag'

    logical_types = [Categorical, CountryCode, Ordinal(order=(1, 2, 3)), PostalCode, SubRegionCode]
    for logical_type in logical_types:
        series = pd.Series([1, 2, 3], dtype='category')
        series.ww.init(logical_type=logical_type, semantic_tags=semantic_tags)
        assert series.ww.semantic_tags == {'custom_tag', 'category'}


def test_does_not_add_standard_tags():
    series = pd.Series([1.1, 2, 3])
    semantic_tags = 'custom_tag'
    series.ww.init(logical_type=Double,
                   semantic_tags=semantic_tags,
                   use_standard_tags=False)
    assert series.ww.semantic_tags == {'custom_tag'}


def test_add_custom_tags(sample_series):
    semantic_tags = 'initial_tag'
    sample_series.ww.init(semantic_tags=semantic_tags)

    sample_series.ww.add_semantic_tags('string_tag')
    assert sample_series.ww.semantic_tags == {'initial_tag', 'string_tag', 'category'}

    sample_series.ww.add_semantic_tags(['list_tag'])
    assert sample_series.ww.semantic_tags == {'initial_tag', 'string_tag', 'list_tag', 'category'}

    sample_series.ww.add_semantic_tags({'set_tag'})
    assert sample_series.ww.semantic_tags == {'initial_tag', 'string_tag', 'list_tag', 'set_tag', 'category'}


def test_set_logical_type_without_standard_tags(sample_series):
    sample_series.ww.init(logical_type='Categorical',
                          semantic_tags='original_tag',
                          use_standard_tags=False)

    new_series = sample_series.ww.set_logical_type('CountryCode')
    assert isinstance(sample_series.ww.logical_type, Categorical)
    assert isinstance(new_series.ww.logical_type, CountryCode)
    assert new_series.ww.semantic_tags == set()


def test_set_logical_type_valid_dtype_change(sample_series):
    sample_series.ww.init(logical_type='Categorical')

    new_series = sample_series.ww.set_logical_type('NaturalLanguage')

    if _is_koalas_series(sample_series):
        # Koalas uses string dtype for Categorical
        original_dtype = 'string'
    else:
        original_dtype = 'category'
    new_dtype = 'string'

    assert isinstance(sample_series.ww.logical_type, Categorical)
    assert sample_series.dtype == original_dtype
    assert isinstance(new_series.ww.logical_type, NaturalLanguage)
    assert new_series.dtype == new_dtype


def test_set_logical_type_invalid_dtype_change(sample_series):
    if _is_dask_series(sample_series):
        pytest.xfail('Dask type conversion with astype does not fail until compute is called')
    if _is_koalas_series(sample_series):
        pytest.xfail('Koalas allows this conversion, filling values it cannot convert with NaN '
                     'and converting dtype to float.')
    sample_series.ww.init(logical_type='Categorical')
    error_message = "Error converting datatype for sample_series from type category to " \
        "type int64. Please confirm the underlying data is consistent with logical type Integer."
    with pytest.raises(TypeConversionError, match=error_message):
        sample_series.ww.set_logical_type('Integer')


def test_reset_semantic_tags_without_standard_tags(sample_series):
    semantic_tags = 'initial_tag'
    sample_series.ww.init(semantic_tags=semantic_tags, use_standard_tags=False)

    sample_series.ww.reset_semantic_tags()
    assert sample_series.ww.semantic_tags == set()


def test_remove_semantic_tags(sample_series):
    tags_to_remove = [
        'tag1',
        ['tag1'],
        {'tag1'}
    ]

    for tag in tags_to_remove:
        series = sample_series.copy()
        series.ww.init(semantic_tags=['tag1', 'tag2'], use_standard_tags=True)
        series.ww.remove_semantic_tags(tag)
        assert series.ww.semantic_tags == {'tag2', 'category'}


def test_series_methods_on_accessor(sample_series):
    sample_series.ww.init()

    copied_series = sample_series.ww.copy()
    assert copied_series is not sample_series
    assert copied_series.ww.schema == sample_series.ww.schema
    pd.testing.assert_series_equal(to_pandas(sample_series), to_pandas(copied_series))


def test_series_methods_on_accessor_without_standard_tags(sample_series):
    sample_series.ww.init(use_standard_tags=False)

    copied_series = sample_series.ww.copy()
    assert copied_series is not sample_series
    assert copied_series.ww.schema == sample_series.ww.schema
    pd.testing.assert_series_equal(to_pandas(sample_series), to_pandas(copied_series))


def test_series_methods_on_accessor_returning_series_valid_schema(sample_series):
    if _is_koalas_series(sample_series):
        pytest.xfail('Running replace on Koalas series changes series dtype to object, invalidating schema')
    sample_series.ww.init()

    replace_series = sample_series.ww.replace('a', 'd')
    assert replace_series.ww._schema == sample_series.ww._schema
    assert replace_series.ww._schema is not sample_series.ww._schema
    pd.testing.assert_series_equal(to_pandas(replace_series), to_pandas(sample_series.replace('a', 'd')))


def test_series_methods_on_accessor_dtype_mismatch(sample_df):
    ints_series = sample_df['id']
    ints_series.ww.init()

    assert isinstance(ints_series.ww.logical_type, Integer)
    assert str(ints_series.dtype) == 'int64'

    warning = ("Operation performed by astype has invalidated the Woodwork typing information:\n "
               "dtype mismatch between original dtype, int64, and returned dtype, Int64.\n "
               "Please initialize Woodwork with Series.ww.init")
    with pytest.warns(TypingInfoMismatchWarning, match=warning):
        series = ints_series.ww.astype('Int64')
    assert series.ww._schema is None


def test_series_methods_on_accessor_inplace(sample_series):
    # TODO: Try to find a supported inplace method for Dask, if one exists
    if _is_dask_series(sample_series):
        pytest.xfail('Dask does not support pop.')
    comparison_series = sample_series.copy()

    sample_series.ww.init()
    comparison_series.ww.init()

    val = sample_series.ww.pop(0)
    assert sample_series.ww._schema == comparison_series.ww._schema
    assert len(sample_series) == len(comparison_series) - 1
    assert val == 'a'


def test_series_methods_on_accessor_returning_series_invalid_schema(sample_series):
    sample_series.ww.init()

    if _is_koalas_series(sample_series):
        # Koalas uses `string` for Categorical, so must try a different conversion
        original_type = 'string'
        new_type = 'Int64'
    else:
        original_type = 'category'
        new_type = 'string'

    warning = "Operation performed by astype has invalidated the Woodwork typing information:\n " \
        f"dtype mismatch between original dtype, {original_type}, and returned dtype, {new_type}.\n " \
        "Please initialize Woodwork with Series.ww.init"

    with pytest.warns(TypingInfoMismatchWarning, match=warning):
        new_series = sample_series.ww.astype(new_type)

    assert new_series.ww._schema is None


def test_series_methods_on_accessor_other_returns(sample_series):
    sample_series.ww.init()
    col_shape = sample_series.ww.shape
    series_shape = sample_series.shape
    if _is_dask_series(sample_series):
        col_shape = (col_shape[0].compute(),)
        series_shape = (series_shape[0].compute())
    assert col_shape == (len(sample_series),)
    assert col_shape == series_shape

    assert sample_series.name == sample_series.ww.name
    series_nunique = sample_series.nunique()
    ww_nunique = sample_series.ww.nunique()
    if _is_dask_series(sample_series):
        series_nunique = series_nunique.compute()
        ww_nunique = ww_nunique.compute()
    assert series_nunique == ww_nunique


def test_series_methods_on_accessor_new_schema_object(sample_series):
    sample_series.ww.init(semantic_tags=['new_tag', 'tag2'], metadata={'important_keys': [1, 2, 3]})

    copied_series = sample_series.ww.copy()

    assert copied_series.ww._schema == sample_series.ww._schema
    assert copied_series.ww._schema is not sample_series.ww._schema

    copied_series.ww.metadata['important_keys'].append(4)
    assert copied_series.ww.metadata['important_keys'] == [1, 2, 3, 4]
    assert sample_series.ww.metadata['important_keys'] == [1, 2, 3]

    copied_series.ww.add_semantic_tags(['tag3'])
    assert copied_series.ww.semantic_tags == {'category', 'new_tag', 'tag2', 'tag3'}
    assert sample_series.ww.semantic_tags == {'category', 'new_tag', 'tag2'}


def test_series_getattr_errors(sample_series):
    error_message = "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    with pytest.raises(WoodworkNotInitError, match=error_message):
        sample_series.ww.shape

    sample_series.ww.init()
    error_message = "Woodwork has no attribute 'invalid_attr'"
    with pytest.raises(AttributeError, match=error_message):
        sample_series.ww.invalid_attr


def test_ordinal_requires_instance_on_init(sample_series):
    error_msg = "Must use an Ordinal instance with order values defined"
    with pytest.raises(TypeError, match=error_msg):
        sample_series.ww.init(logical_type=Ordinal)
    with pytest.raises(TypeError, match=error_msg):
        sample_series.ww.init(logical_type="Ordinal")


def test_ordinal_requires_instance_on_update(sample_series):
    sample_series.ww.init(logical_type="Categorical")

    error_msg = "Must use an Ordinal instance with order values defined"
    with pytest.raises(TypeError, match=error_msg):
        sample_series.ww.set_logical_type(Ordinal)
    with pytest.raises(TypeError, match=error_msg):
        sample_series.ww.set_logical_type("Ordinal")


def test_ordinal_with_order(sample_series):
    if _is_koalas_series(sample_series) or _is_dask_series(sample_series):
        pytest.xfail('Fails with Dask and Koalas - ordinal data validation not compatible')

    series = sample_series.copy()
    ordinal_with_order = Ordinal(order=['a', 'b', 'c'])
    series.ww.init(logical_type=ordinal_with_order)
    assert isinstance(series.ww.logical_type, Ordinal)
    assert series.ww.logical_type.order == ['a', 'b', 'c']

    series = sample_series.copy()
    series.ww.init(logical_type='Categorical')
    new_series = series.ww.set_logical_type(ordinal_with_order)
    assert isinstance(new_series.ww.logical_type, Ordinal)
    assert new_series.ww.logical_type.order == ['a', 'b', 'c']


def test_ordinal_with_incomplete_ranking(sample_series):
    if _is_koalas_series(sample_series) or _is_dask_series(sample_series):
        pytest.xfail('Fails with Dask and Koalas - ordinal data validation not supported')

    ordinal_incomplete_order = Ordinal(order=['a', 'b'])
    error_msg = re.escape("Ordinal column sample_series contains values that are not "
                          "present in the order values provided: ['c']")
    with pytest.raises(ValueError, match=error_msg):
        sample_series.ww.init(logical_type=ordinal_incomplete_order)


def test_ordinal_with_nan_values():
    nan_series = pd.Series(['a', 'b', np.nan, 'a']).astype('category')
    ordinal_with_order = Ordinal(order=['a', 'b'])
    nan_series.ww.init(logical_type=ordinal_with_order)
    assert isinstance(nan_series.ww.logical_type, Ordinal)
    assert nan_series.ww.logical_type.order == ['a', 'b']


def test_latlong_init_with_valid_series(latlongs):
    series = latlongs[0]
    series.ww.init(logical_type="LatLong")
    assert isinstance(series.ww.logical_type, LatLong)


def test_latlong_init_error_with_invalid_series(latlongs):
    series = latlongs[1]
    error_message = "Cannot initialize Woodwork. Series does not contain properly formatted " \
        "LatLong data. Try reformatting before initializing or use the " \
        "woodwork.init_series function to initialize."
    with pytest.raises(ValueError, match=error_message):
        series.ww.init(logical_type="LatLong")


def test_latlong_formatting_with_init_series(latlongs):
    expected_series = pd.Series([(1.0, 2.0), (3.0, 4.0)])
    if _is_dask_series(latlongs[0]):
        expected_series = dd.from_pandas(expected_series, npartitions=2)
    elif _is_koalas_series(latlongs[0]):
        expected_series = ks.Series([[1.0, 2.0], [3.0, 4.0]])

    expected_series.ww.init(logical_type=LatLong)
    for series in latlongs:
        new_series = init_series(series, logical_type=LatLong)
        assert isinstance(new_series.ww.logical_type, LatLong)
        pd.testing.assert_series_equal(to_pandas(new_series), to_pandas(expected_series))
        assert expected_series.ww._schema == new_series.ww._schema


def test_accessor_equality(sample_series):
    # Check different parameters
    str_col = sample_series.copy()
    str_col.ww.init(logical_type='Categorical')

    str_col_2 = sample_series.copy()
    str_col_2.ww.init(logical_type=Categorical)

    str_col_diff_tags = sample_series.copy()
    str_col_diff_tags.ww.init(logical_type=Categorical, semantic_tags={'test'})

    diff_name_col = sample_series.copy()
    diff_name_col.name = 'different_name'
    diff_name_col.ww.init(logical_type=Categorical)

    diff_dtype_col = sample_series.astype('string')
    diff_dtype_col.ww.init(logical_type=NaturalLanguage)

    assert str_col.ww == str_col_2.ww
    assert str_col.ww != str_col_diff_tags.ww
    assert str_col.ww != diff_name_col.ww
    assert str_col.ww != diff_dtype_col.ww

    # Check different underlying series
    str_col = sample_series.astype('string')
    str_col.ww.init(logical_type='NaturalLanguage')
    changed_series = sample_series.copy().replace(to_replace='a', value='test')
    changed_series = changed_series.astype('string')
    changed_series.ww.init(logical_type='NaturalLanguage')

    # We only check underlying data for equality with pandas dataframes
    if isinstance(str_col, pd.Series):
        assert str_col.ww != changed_series.ww
    else:
        assert str_col.ww == changed_series.ww


def test_accessor_shallow_equality(sample_series):
    metadata_col = init_series(sample_series.copy(), logical_type='NaturalLanguage',
                               metadata={'interesting_values': ['a', 'b']})
    diff_metadata_col = init_series(sample_series.copy(), logical_type='NaturalLanguage',
                                    metadata={'interesting_values': ['c']})

    assert metadata_col.ww.__eq__(diff_metadata_col.ww, deep=False)
    assert not metadata_col.ww.__eq__(diff_metadata_col.ww, deep=True)

    schema = metadata_col.ww.schema
    diff_data_col = metadata_col.replace({'a': '1'})
    # dtype gets changed to object in replace
    diff_data_col = diff_data_col.astype('string')

    diff_data_col.ww.init(schema=schema)
    same_data_col = metadata_col.ww.copy()

    assert diff_data_col.ww.schema.__eq__(metadata_col.ww.schema, deep=True)
    assert same_data_col.ww.schema.__eq__(metadata_col.ww.schema, deep=True)

    assert diff_data_col.ww.__eq__(metadata_col.ww, deep=False)
    assert same_data_col.ww.__eq__(metadata_col.ww, deep=False)
    assert same_data_col.ww.__eq__(metadata_col.ww, deep=True)
    if isinstance(sample_series, pd.Series):
        # We only check underlying data for equality with pandas dataframes
        assert not diff_data_col.ww.__eq__(metadata_col.ww, deep=True)
    else:
        assert diff_data_col.ww.__eq__(metadata_col.ww, deep=True)


def test_accessor_metadata(sample_series):
    column_metadata = {'metadata_field': [1, 2, 3], 'created_by': 'user0'}

    series = sample_series.copy()
    series.ww.init()
    assert series.ww.metadata == {}

    series = sample_series.copy()
    series.ww.init(metadata=column_metadata)
    assert series.ww.metadata == column_metadata

    new_metadata = {'date_created': '1/1/19', 'created_by': 'user1'}

    series.ww.metadata = {**series.ww.metadata, **new_metadata}
    assert series.ww.metadata == {'date_created': '1/1/19', 'metadata_field': [1, 2, 3], 'created_by': 'user1'}

    series.ww.metadata.pop('created_by')
    assert series.ww.metadata == {'date_created': '1/1/19', 'metadata_field': [1, 2, 3]}

    series.ww.metadata['number'] = 1012034
    assert series.ww.metadata == {'date_created': '1/1/19', 'metadata_field': [1, 2, 3], 'number': 1012034}


def test_metadata_setter_error_before_init(sample_series):
    err_msg = "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    with pytest.raises(WoodworkNotInitError, match=err_msg):
        sample_series.ww.metadata = {"key": "val"}


def test_accessor_metadata_error_on_init(sample_series):
    err_msg = "Column metadata must be a dictionary"
    with pytest.raises(TypeError, match=err_msg):
        sample_series.ww.init(metadata=123)


def test_accessor_metadata_error_on_update(sample_series):
    sample_series.ww.init()
    err_msg = "Column metadata must be a dictionary"
    with pytest.raises(TypeError, match=err_msg):
        sample_series.ww.metadata = 123


def test_non_string_column_name(sample_series):
    sample_series.name = 0
    sample_series.ww.init(semantic_tags={'test_tag'})

    assert sample_series.ww.name == 0
    assert sample_series.name == 0
    assert sample_series.ww.semantic_tags == {'category', 'test_tag'}


def test_schema_property(sample_series):
    sample_series.ww.init(logical_type='Categorical', semantic_tags='test_tag', metadata={'created_by': 'user1'})

    assert sample_series.ww.schema is not sample_series.ww._schema
    assert sample_series.ww.schema == sample_series.ww._schema

    changed_schema = sample_series.ww.schema
    changed_schema.metadata['new'] = 1

    assert changed_schema != sample_series.ww._schema


@patch("woodwork.column_accessor.WoodworkColumnAccessor._validate_logical_type")
def test_validation_methods_called_init(mock_validate_logical_type, sample_series):
    assert not mock_validate_logical_type.called

    not_validated = sample_series.copy()
    not_validated.ww.init(validate=False)

    assert not mock_validate_logical_type.called

    validated = sample_series.copy()
    validated.ww.init(validate=True)

    assert mock_validate_logical_type.called
    assert validated.ww == not_validated.ww
    pd.testing.assert_series_equal(to_pandas(validated), to_pandas(not_validated))


@patch("woodwork.column_accessor._validate_schema")
def test_validation_methods_called_init_with_schema(mock_validate_schema, sample_series):
    assert not mock_validate_schema.called
    schema_series = sample_series.copy()
    schema_series.ww.init()
    schema = schema_series.ww.schema

    not_validated = sample_series.copy()
    not_validated.ww.init(schema=schema, validate=False)

    assert not mock_validate_schema.called

    validated = sample_series.copy()
    validated.ww.init(schema=schema, validate=True)

    assert mock_validate_schema.called
    assert validated.ww == not_validated.ww
    pd.testing.assert_series_equal(to_pandas(validated), to_pandas(not_validated))


def test_series_methods_returning_frame(sample_series):
    sample_series.ww.init(semantic_tags={'test_tag'},
                          description='custom description',
                          origin='base',
                          metadata={'custom key': 'custom value'})
    sample_frame = sample_series.ww.to_frame()

    assert _is_dataframe(sample_frame)
    assert sample_frame.ww.schema is not None
    assert sample_frame.ww.columns['sample_series'] == sample_series.ww.schema

    reset_index_frame = sample_series.ww.reset_index(drop=False)
    assert _is_dataframe(reset_index_frame)
    assert reset_index_frame.ww.schema is None



import numpy as np
import pandas as pd

from woodwork.utils import import_or_none

dd = import_or_none('dask.dataframe')
ks = import_or_none('databricks.koalas')


def validate_subset_schema(subset_schema, schema):
    assert subset_schema.name == schema.name
    for subset_col_name, subset_col in subset_schema.columns.items():
        assert subset_col_name in schema.columns
        col = schema.columns[subset_col_name]
        assert subset_col.logical_type == col.logical_type
        assert subset_col.semantic_tags == col.semantic_tags


def mi_between_cols(col1, col2, df):
    mi_series = df.loc[df['column_1'] == col1].loc[df['column_2'] == col2]['mutual_info']

    if len(mi_series) == 0:
        mi_series = df.loc[df['column_1'] == col2].loc[df['column_2'] == col1]['mutual_info']

    return mi_series.iloc[0]


def to_pandas(df, index=None, sort_index=False):
    """Testing util to convert dataframes to pandas. If a pandas dataframe is passed in, just returns the dataframe.

    Returns:
        Pandas DataFrame
    """
    if isinstance(df, (pd.DataFrame, pd.Series, pd.Index)):
        return df

    if dd and isinstance(df, (dd.DataFrame, dd.Series, dd.Index)):
        pd_df = df.compute()

    if ks and isinstance(df, (ks.DataFrame, ks.Series, ks.Index)):
        pd_df = df.to_pandas()

    if index:
        pd_df = pd_df.set_index(index, drop=False)
    if sort_index:
        pd_df = pd_df.sort_index()

    return pd_df


def is_public_method(class_to_check, name):
    """Determine if the specified name is a public method on a class"""
    if hasattr(class_to_check, name) and name[0] != '_':
        if callable(getattr(class_to_check, name)):
            return True
    return False


def is_property(class_to_check, name):
    """Determine if the specified name is a property on a class"""
    if hasattr(class_to_check, name) and isinstance(getattr(class_to_check, name), property):
        return True
    return False


def check_empty_box_plot_dict(box_plot_dict):
    assert np.isnan(box_plot_dict['low_bound'])
    assert np.isnan(box_plot_dict['high_bound'])
    assert len(box_plot_dict['quantiles']) == 5
    assert all([np.isnan(elt) for elt in box_plot_dict['quantiles'].values()])
    assert len(box_plot_dict['low_values']) == 0
    assert len(box_plot_dict['high_values']) == 0



import pandas as pd
import pytest

import woodwork as ww
from woodwork.logical_types import (
    Boolean,
    Categorical,
    CountryCode,
    Datetime,
    Double,
    Integer,
    LogicalType,
    NaturalLanguage,
    Ordinal,
    PersonFullName,
    SubRegionCode,
    Unknown
)
from woodwork.type_sys.inference_functions import (
    categorical_func,
    double_func,
    integer_func
)
from woodwork.type_sys.type_system import TypeSystem


def test_type_system_init(default_inference_functions, default_relationships):
    type_sys = TypeSystem(inference_functions=default_inference_functions,
                          relationships=default_relationships)
    assert len(type_sys.inference_functions) == 5
    assert type_sys.inference_functions[Double] is double_func
    assert type_sys.inference_functions[Integer] is integer_func
    assert type_sys.inference_functions[Categorical] is categorical_func
    assert len(type_sys.relationships) == 2
    assert type_sys.relationships[0] == (Double, Integer)
    assert type_sys.relationships[1] == (Categorical, CountryCode)


def test_add_type_validation_errors(type_sys):
    error_msg = 'logical_type must be a valid LogicalType'
    with pytest.raises(TypeError, match=error_msg):
        type_sys.add_type(logical_type=1)

    with pytest.raises(TypeError, match=error_msg):
        type_sys.add_type(logical_type=Double())

    error_msg = 'inference_function must be a function'
    with pytest.raises(TypeError, match=error_msg):
        type_sys.add_type(logical_type=Ordinal, inference_function='not a function')

    error_msg = 'parent must be a valid LogicalType'
    with pytest.raises(ValueError, match=error_msg):
        type_sys.add_type(logical_type=Ordinal, parent=1)

    with pytest.raises(ValueError, match=error_msg):
        type_sys.add_type(logical_type=Ordinal, parent=Double())


def test_remove_type_validation_errors(type_sys):
    error_msg = 'logical_type must be a valid LogicalType'
    with pytest.raises(TypeError, match=error_msg):
        type_sys.add_type(logical_type=1)


def test_update_inference_function_validation_errors(type_sys):
    error_msg = 'logical_type must be a valid LogicalType'
    with pytest.raises(TypeError, match=error_msg):
        type_sys.update_inference_function(logical_type=1, inference_function=None)

    error_msg = 'inference_function must be a function'
    with pytest.raises(TypeError, match=error_msg):
        type_sys.update_inference_function(logical_type=Ordinal, inference_function='not a function')


def test_update_relationship_validation_errors(type_sys):
    error_msg = 'logical_type must be a valid LogicalType'
    with pytest.raises(TypeError, match=error_msg):
        type_sys.update_relationship(logical_type=1, parent=Ordinal)

    error_msg = 'parent must be a valid LogicalType'
    with pytest.raises(ValueError, match=error_msg):
        type_sys.update_relationship(logical_type=Ordinal, parent=1)


def test_type_system_default_type(default_inference_functions, default_relationships):
    type_sys = TypeSystem(inference_functions=default_inference_functions,
                          relationships=default_relationships,
                          default_type=SubRegionCode)
    assert type_sys.default_type == SubRegionCode
    type_sys.update_inference_function(Categorical, None)
    test_series = pd.Series(['a', 'b', 'c'])
    assert isinstance(type_sys.infer_logical_type(test_series), SubRegionCode)
    assert SubRegionCode in type_sys.registered_types


def test_type_system_default_type_remove_error(default_inference_functions, default_relationships):
    type_sys = TypeSystem(inference_functions=default_inference_functions,
                          relationships=default_relationships,
                          default_type=SubRegionCode)
    error_msg = "Default LogicalType cannot be removed"
    with pytest.raises(ValueError, match=error_msg):
        type_sys.remove_type(SubRegionCode)
    with pytest.raises(ValueError, match=error_msg):
        type_sys.remove_type('SubRegionCode')
    with pytest.raises(ValueError, match=error_msg):
        type_sys.remove_type('sub_region_code')


def test_type_system_registered_types(type_sys):
    assert isinstance(type_sys.registered_types, list)
    assert set(type_sys.registered_types) == {Double, Integer, Categorical, CountryCode, Unknown}


def test_type_system_root_types(type_sys):
    assert isinstance(type_sys.root_types, list)
    assert set(type_sys.root_types) == {Double, Categorical, Unknown}


def test_add_type_without_parent():
    type_sys = TypeSystem()
    type_sys.add_type(Double, inference_function=double_func)
    assert Double in type_sys.inference_functions.keys()
    assert type_sys.inference_functions[Double] is double_func

    type_sys.add_type(Integer)
    assert Integer in type_sys.inference_functions.keys()
    assert type_sys.inference_functions[Integer] is None


def test_add_type_with_parent():
    type_sys = TypeSystem()
    type_sys.add_type(Double, inference_function=double_func)
    type_sys.add_type(Integer, inference_function=integer_func, parent=Double)
    assert Integer in type_sys.inference_functions.keys()
    assert type_sys.inference_functions[Integer] is integer_func
    assert len(type_sys.relationships) == 1
    assert type_sys.relationships[0] == (Double, Integer)


def test_add_duplicate_ltype(type_sys):
    inference_fn = type_sys.inference_functions[ww.logical_types.Integer]

    assert ww.type_system.str_to_logical_type('Integer') == ww.logical_types.Integer

    class Integer(LogicalType):
        primary_dtype = 'string'

    error_msg = 'Logical Type with name Integer already present in the Type System. Please rename the LogicalType or remove existing one.'
    with pytest.raises(ValueError, match=error_msg):
        type_sys.add_type(Integer, inference_function=inference_fn)

    type_sys.remove_type(ww.logical_types.Integer)
    type_sys.add_type(Integer, inference_function=inference_fn)

    ltype = type_sys.str_to_logical_type('Integer')
    assert ltype.primary_dtype == 'string'
    assert ltype == Integer

    type_sys.reset_defaults()
    ltype = type_sys.str_to_logical_type('Integer')
    assert ltype.primary_dtype == 'int64'
    assert ltype == ww.logical_types.Integer


def test_remove_type_no_children(type_sys):
    type_sys.remove_type(Integer)
    assert len(type_sys.inference_functions) == 4
    assert Integer not in type_sys.inference_functions.keys()
    assert len(type_sys.relationships) == 1
    type_sys.remove_type('CountryCode')
    assert len(type_sys.inference_functions) == 3
    assert CountryCode not in type_sys.inference_functions.keys()
    assert len(type_sys.relationships) == 0


def test_remove_type_with_children(type_sys):
    type_sys.remove_type(Double)
    assert len(type_sys.inference_functions) == 4
    assert Double not in type_sys.inference_functions.keys()
    assert len(type_sys.relationships) == 2
    assert (None, Integer) in type_sys.relationships
    assert Integer in type_sys.root_types


def test_get_children(type_sys):
    type_sys.add_type(Ordinal, parent=Categorical)
    children = type_sys._get_children(Categorical)
    assert isinstance(children, list)
    assert len(children) == 2
    assert set(children) == {CountryCode, Ordinal}


def test_get_parent(type_sys):
    assert type_sys._get_parent(CountryCode) == Categorical
    assert type_sys._get_parent(Integer) == Double


def test_get_depth(type_sys):
    type_sys.add_type(Ordinal, parent=CountryCode)
    assert type_sys._get_depth(Categorical) == 0
    assert type_sys._get_depth(CountryCode) == 1
    assert type_sys._get_depth(Ordinal) == 2


def test_update_inference_function(type_sys):
    assert type_sys.inference_functions[Double] is double_func
    type_sys.update_inference_function(Double, integer_func)
    assert type_sys.inference_functions[Double] is integer_func
    type_sys.update_inference_function('CountryCode', integer_func)
    assert type_sys.inference_functions[CountryCode] is integer_func


def test_update_relationship_no_children(type_sys):
    type_sys.update_relationship(CountryCode, Integer)
    assert len(type_sys.relationships) == 2
    assert (Integer, CountryCode) in type_sys.relationships
    assert type_sys._get_parent(CountryCode) == Integer


def test_update_relationship_string_input(type_sys):
    type_sys.update_relationship('CountryCode', 'Integer')
    assert len(type_sys.relationships) == 2
    assert (Integer, CountryCode) in type_sys.relationships
    assert type_sys._get_parent(CountryCode) == Integer


def test_update_relationship_with_children(type_sys):
    type_sys.add_type(SubRegionCode, parent=CountryCode)
    type_sys.update_relationship(CountryCode, Integer)
    assert len(type_sys.relationships) == 3
    assert (Integer, CountryCode) in type_sys.relationships
    assert (CountryCode, SubRegionCode) in type_sys.relationships
    assert type_sys._get_children(CountryCode) == [SubRegionCode]


def test_inference_multiple_matches_same_depth(default_relationships):
    def always_true(series):
        return True
    inference_functions = {
        Categorical: always_true,
        Double: always_true,
        Integer: always_true,
        CountryCode: always_true,
    }
    type_sys = TypeSystem(inference_functions=inference_functions,
                          relationships=default_relationships,
                          default_type=NaturalLanguage)
    type_sys.update_inference_function(Integer, always_true)
    type_sys.update_inference_function(CountryCode, always_true)
    inferred_type = type_sys.infer_logical_type(pd.Series([1, 2, 3]))
    # Should match CountryCode - same depth as Integer, but CountryCode parent
    # (Categorical) is tried and found first
    assert isinstance(inferred_type, CountryCode)


def test_inference_multiple_matches_different_depths(default_relationships):
    def always_true(series):
        return True
    inference_functions = {
        Categorical: always_true,
        Double: always_true,
        Integer: always_true,
        CountryCode: always_true,
    }
    type_sys = TypeSystem(inference_functions=inference_functions,
                          relationships=default_relationships,
                          default_type=NaturalLanguage)
    type_sys.update_inference_function(Integer, always_true)
    type_sys.update_inference_function(CountryCode, always_true)
    type_sys.add_type(SubRegionCode, inference_function=always_true, parent=CountryCode)
    inferred_type = type_sys.infer_logical_type(pd.Series([1, 2, 3]))
    # Should match SubRegionCode as it is the deepest match
    assert isinstance(inferred_type, SubRegionCode)


def test_reset_defaults(type_sys, default_inference_functions, default_relationships):
    type_sys.update_inference_function('Integer', None)
    type_sys.update_relationship(CountryCode, parent=Unknown)
    type_sys.default_type = Categorical
    type_sys.reset_defaults()

    assert type_sys.inference_functions == default_inference_functions
    assert type_sys.relationships == default_relationships
    assert type_sys.default_type == Unknown


def test_get_logical_types():
    all_types = ww.type_system.registered_types
    logical_types = ww.type_system._get_logical_types()

    for logical_type in all_types:
        assert logical_types[logical_type.__name__] == logical_type
        assert logical_types[logical_type.type_string] == logical_type

    assert len(logical_types) == 2 * len(all_types)


def test_str_to_logical_type():
    all_types = ww.type_system.registered_types

    with pytest.raises(ValueError, match='String test is not a valid logical type'):
        ww.type_system.str_to_logical_type('test')
    assert ww.type_system.str_to_logical_type('test', raise_error=False) is None

    for logical_type in all_types:
        assert ww.type_system.str_to_logical_type(logical_type.__name__) == logical_type
        assert ww.type_system.str_to_logical_type(logical_type.type_string) == logical_type

    assert ww.type_system.str_to_logical_type('bOoLeAn') == Boolean
    assert ww.type_system.str_to_logical_type('person_full_NAME') == PersonFullName
    assert ww.type_system.str_to_logical_type('PersonFullnamE') == PersonFullName

    ymd = '%Y-%m-%d'
    datetime_with_format = ww.type_system.str_to_logical_type('datetime', params={'datetime_format': ymd})
    assert datetime_with_format.__class__ == Datetime
    assert datetime_with_format.datetime_format == ymd
    assert datetime_with_format == Datetime(datetime_format=ymd)

    datetime_no_format = ww.type_system.str_to_logical_type('datetime', params={'datetime_format': None})
    assert datetime_no_format.__class__ == Datetime
    assert datetime_no_format.datetime_format is None
    assert datetime_no_format == Datetime()

    # When parameters are supplied in a non-empty dictionary, the logical type gets instantiated
    assert ww.type_system.str_to_logical_type('person_full_NAME', params={}) == PersonFullName
    assert datetime_no_format != Datetime

    # Input a different type system
    new_type_sys = TypeSystem()
    with pytest.raises(ValueError, match='String Integer is not a valid logical type'):
        new_type_sys.str_to_logical_type('Integer')
    new_type_sys.add_type(Boolean)
    assert Boolean == new_type_sys.str_to_logical_type('Boolean')



import datetime
import json
import os
import tarfile
import tempfile

import pandas as pd

import woodwork as ww
from woodwork.accessor_utils import _is_dask_dataframe, _is_koalas_dataframe
from woodwork.s3_utils import get_transport_params, use_smartopen
from woodwork.type_sys.utils import (
    _get_ltype_class,
    _get_specified_ltype_params
)
from woodwork.utils import _is_s3, _is_url

SCHEMA_VERSION = '11.1.0'
FORMATS = ['csv', 'pickle', 'parquet', 'arrow', 'feather', 'orc']


def typing_info_to_dict(dataframe):
    """Creates the description for a Woodwork table, including typing information for each column
    and loading information.

    Args:
        dataframe (pd.DataFrame, dd.Dataframe, ks.DataFrame): DataFrame with Woodwork typing
            information initialized.

    Returns:
        dict: Dictionary containing Woodwork typing information
    """
    if _is_dask_dataframe(dataframe):
        # Need to determine the category info for Dask it can be saved below
        category_cols = [colname for colname, col in dataframe.ww._schema.columns.items() if col.is_categorical]
        dataframe = dataframe.ww.categorize(columns=category_cols)
    ordered_columns = dataframe.columns

    def _get_physical_type_dict(column):
        type_dict = {'type': str(column.dtype)}
        if str(column.dtype) == 'category':
            type_dict['cat_values'] = column.dtype.categories.to_list()
            type_dict['cat_dtype'] = str(column.dtype.categories.dtype)
        return type_dict

    column_typing_info = [
        {'name': col_name,
         'ordinal': ordered_columns.get_loc(col_name),
         'use_standard_tags': col.use_standard_tags,
         'logical_type': {
             'parameters': _get_specified_ltype_params(col.logical_type),
             'type': str(_get_ltype_class(col.logical_type))
         },
         'physical_type': _get_physical_type_dict(dataframe[col_name]),
         'semantic_tags': sorted(list(col.semantic_tags)),
         'description': col.description,
         'origin': col.origin,
         'metadata': col.metadata,
         }
        for col_name, col in dataframe.ww.columns.items()
    ]

    if _is_dask_dataframe(dataframe):
        table_type = 'dask'
    elif _is_koalas_dataframe(dataframe):
        table_type = 'koalas'
    else:
        table_type = 'pandas'

    return {
        'schema_version': SCHEMA_VERSION,
        'name': dataframe.ww.name,
        'index': dataframe.ww.index,
        'time_index': dataframe.ww.time_index,
        'column_typing_info': column_typing_info,
        'loading_info': {
            'table_type': table_type
        },
        'table_metadata': dataframe.ww.metadata
    }


def write_woodwork_table(dataframe, path, profile_name=None, **kwargs):
    """Serialize Woodwork table and write to disk or S3 path.

    Args:
        dataframe (pd.DataFrame, dd.DataFrame, ks.DataFrame): DataFrame with Woodwork typing information initialized.
        path (str) : Location on disk to write the Woodwork table.
        profile_name (str, bool): The AWS profile specified to write to S3. Will default to None and search for AWS credentials.
                Set to False to use an anonymous profile.
        kwargs (keywords) : Additional keyword arguments to pass as keywords arguments to the underlying serialization method or to specify AWS profile.
    """
    if _is_s3(path):
        with tempfile.TemporaryDirectory() as tmpdir:
            os.makedirs(os.path.join(tmpdir, 'data'))
            _dump_table(dataframe, tmpdir, **kwargs)
            file_path = _create_archive(tmpdir)

            transport_params = get_transport_params(profile_name)
            use_smartopen(file_path, path, read=False, transport_params=transport_params)
    elif _is_url(path):
        raise ValueError("Writing to URLs is not supported")
    else:
        path = os.path.abspath(path)
        os.makedirs(os.path.join(path, 'data'), exist_ok=True)
        _dump_table(dataframe, path, **kwargs)


def _dump_table(dataframe, path, **kwargs):
    """Writes Woodwork table at the specified path, including both the data and the typing information."""
    loading_info = write_dataframe(dataframe, path, **kwargs)

    typing_info = typing_info_to_dict(dataframe)
    typing_info['loading_info'].update(loading_info)

    write_typing_info(typing_info, path)


def write_typing_info(typing_info, path):
    """Writes Woodwork typing information to the specified path at woodwork_typing_info.json

    Args:
        typing_info (dict): Dictionary containing Woodwork typing information.
    """
    try:
        file = os.path.join(path, 'woodwork_typing_info.json')
        with open(file, 'w') as file:
            json.dump(typing_info, file)
    except TypeError:
        raise TypeError('Woodwork table is not json serializable. Check table and column metadata for values that may not be serializable.')


def write_dataframe(dataframe, path, format='csv', **kwargs):
    """Write underlying DataFrame data to disk or S3 path.

    Args:
        dataframe (pd.DataFrame, dd.DataFrame, ks.DataFrame): DataFrame with Woodwork typing information initialized.
        path (str) : Location on disk to write the Woodwork table.
        format (str) : Format to use for writing Woodwork data. Defaults to csv.
        kwargs (keywords) : Additional keyword arguments to pass as keywords arguments to the underlying serialization method.

    Returns:
        dict: Information on storage location and format of data.
    """
    format = format.lower()

    ww_name = dataframe.ww.name or 'data'

    if _is_dask_dataframe(dataframe) and format == 'csv':
        basename = "{}-*.{}".format(ww_name, format)
    else:
        basename = '.'.join([ww_name, format])
    location = os.path.join('data', basename)
    file = os.path.join(path, location)

    if format == 'csv':
        # engine kwarg not needed for writing, only reading
        csv_kwargs = kwargs.copy()
        if 'engine' in csv_kwargs.keys():
            del csv_kwargs['engine']
        if _is_koalas_dataframe(dataframe):
            dataframe = dataframe.ww.copy()
            columns = list(dataframe.select_dtypes('object').columns)
            dataframe[columns] = dataframe[columns].astype(str)
            csv_kwargs['compression'] = str(csv_kwargs['compression'])
        dataframe.to_csv(file, **csv_kwargs)
    elif format == 'pickle':
        # Dask and Koalas currently do not support to_pickle
        if not isinstance(dataframe, pd.DataFrame):
            msg = 'DataFrame type not compatible with pickle serialization. Please serialize to another format.'
            raise ValueError(msg)
        dataframe.to_pickle(file, **kwargs)
    elif format in ['parquet', 'arrow', 'feather', 'orc']:
        # Latlong columns in pandas and Dask DataFrames contain tuples, which raises
        # an error in parquet and arrow/feather format.
        latlong_columns = [col_name for col_name, col in dataframe.ww.columns.items() if _get_ltype_class(col.logical_type) == ww.logical_types.LatLong]
        if len(latlong_columns) > 0:
            dataframe = dataframe.ww.copy()
            dataframe[latlong_columns] = dataframe[latlong_columns].astype(str)
        if format == 'parquet':
            dataframe.to_parquet(file, **kwargs)
        elif format == 'orc':
            # Serialization to orc relies on pyarrow.Table.from_pandas which doesn't work with Dask
            if _is_dask_dataframe(dataframe):
                msg = 'DataFrame type not compatible with orc serialization. Please serialize to another format.'
                raise ValueError(msg)
            save_orc_file(dataframe, file)
        else:
            dataframe.to_feather(file, **kwargs)
    else:
        error = 'must be one of the following formats: {}'
        raise ValueError(error.format(', '.join(FORMATS)))
    return {'location': location, 'type': format, 'params': kwargs}


def _create_archive(tmpdir):
    """When seralizing to an S3 URL, writes a tar archive."""
    file_name = "ww-{date:%Y-%m-%d_%H%M%S}.tar".format(date=datetime.datetime.now())
    file_path = os.path.join(tmpdir, file_name)
    tar = tarfile.open(str(file_path), 'w')
    tar.add(str(tmpdir) + '/woodwork_typing_info.json', arcname='/woodwork_typing_info.json')
    tar.add(str(tmpdir) + '/data', arcname='/data')
    tar.close()
    return file_path


def save_orc_file(dataframe, filepath):
    from pyarrow import Table, orc
    df = dataframe.copy()
    for c in df:
        if df[c].dtype.name == 'category':
            df[c] = df[c].astype('string')
    pa_table = Table.from_pandas(df, preserve_index=False)
    orc.write_table(pa_table, filepath)



import woodwork as ww
from woodwork.accessor_utils import _is_koalas_series
from woodwork.logical_types import (
    Boolean,
    BooleanNullable,
    Categorical,
    Datetime,
    Double,
    EmailAddress,
    Integer,
    IntegerNullable,
    LogicalType,
    Timedelta,
    Unknown
)
from woodwork.tests.testing_utils import to_pandas
from woodwork.utils import import_or_none

UNSUPPORTED_KOALAS_DTYPES = [
    'int32',
    'intp',
    'uint8',
    'uint16',
    'uint32',
    'uint64',
    'uintp',
    'float_',
    'object',
    'category',
]

ks = import_or_none('databricks.koalas')


def get_koalas_dtypes(dtypes):
    return [dtype for dtype in dtypes if dtype not in UNSUPPORTED_KOALAS_DTYPES]


def test_integer_inference(integers):
    dtypes = ['int8', 'int16', 'int32', 'int64', 'intp', 'int', 'Int64']
    if _is_koalas_series(integers[0]):
        dtypes = get_koalas_dtypes(dtypes)

    for series in integers:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Integer)


def test_double_inference(doubles):
    dtypes = ['float', 'float32', 'float64', 'float_']
    if _is_koalas_series(doubles[0]):
        dtypes = get_koalas_dtypes(dtypes)

    for series in doubles:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Double)


def test_boolean_inference(bools):
    dtypes = ['bool', 'boolean']

    for series in bools:
        for dtype in dtypes:
            cast_series = series.astype(dtype)
            inferred_type = ww.type_system.infer_logical_type(cast_series)
            if to_pandas(cast_series).isnull().any():
                assert isinstance(inferred_type, BooleanNullable)
            else:
                assert isinstance(inferred_type, Boolean)


def test_datetime_inference(datetimes):
    dtypes = ['object', 'string', 'datetime64[ns]']
    if _is_koalas_series(datetimes[0]):
        dtypes = get_koalas_dtypes(dtypes)

    for series in datetimes:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Datetime)


def test_email_inference(emails):
    dtypes = ['object', 'string']
    if _is_koalas_series(emails[0]):
        dtypes = get_koalas_dtypes(dtypes)

    for series in emails:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, EmailAddress)


def test_email_inference_failure(bad_emails):
    dtypes = ['object', 'string']
    if _is_koalas_series(bad_emails[0]):
        dtypes = get_koalas_dtypes(dtypes)

    for series in bad_emails:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert not isinstance(inferred_type, EmailAddress)


def test_categorical_inference(categories):
    dtypes = ['object', 'string', 'category']
    if _is_koalas_series(categories[0]):
        dtypes = get_koalas_dtypes(dtypes)
    for series in categories:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Categorical)


def test_categorical_inference_based_on_dtype(categories_dtype):
    """
    This test specifically targets the case in which a series can be inferred
    to be categorical strictly from its pandas dtype, but would otherwise be
    inferred as some other type.
    """
    inferred_type = ww.type_system.infer_logical_type(categories_dtype["cat"])
    assert isinstance(inferred_type, Categorical)

    inferred_type = ww.type_system.infer_logical_type(categories_dtype["non_cat"])
    assert not isinstance(inferred_type, Categorical)


def test_categorical_integers_inference(integers):
    with ww.config.with_options(numeric_categorical_threshold=0.5):
        dtypes = ['int8', 'int16', 'int32', 'int64', 'intp', 'int', 'Int64']
        if _is_koalas_series(integers[0]):
            dtypes = get_koalas_dtypes(dtypes)
        for series in integers:
            for dtype in dtypes:
                inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
                assert isinstance(inferred_type, Categorical)


def test_categorical_double_inference(doubles):
    with ww.config.with_options(numeric_categorical_threshold=0.5):
        dtypes = ['float', 'float32', 'float64', 'float_']
        if _is_koalas_series(doubles[0]):
            dtypes = get_koalas_dtypes(dtypes)
        for series in doubles:
            for dtype in dtypes:
                inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
                assert isinstance(inferred_type, Categorical)


def test_timedelta_inference(timedeltas):
    dtypes = ['timedelta64[ns]']
    for series in timedeltas:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Timedelta)


def test_unknown_inference(strings):
    dtypes = ['object', 'string']
    if _is_koalas_series(strings[0]):
        dtypes = get_koalas_dtypes(dtypes)

    for series in strings:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Unknown)


def test_unknown_inference_all_null(nulls):
    dtypes = ['object', 'string', 'category', 'datetime64[ns]']
    if _is_koalas_series(nulls[0]):
        dtypes = get_koalas_dtypes(dtypes)

    for series in nulls:
        for dtype in dtypes:
            inferred_type = ww.type_system.infer_logical_type(series.astype(dtype))
            inferred_type.transform(series)
            assert isinstance(inferred_type, Unknown)


def test_pdna_inference(pdnas):
    expected_logical_types = [
        Unknown,
        IntegerNullable,
        BooleanNullable,
    ]

    for index, series in enumerate(pdnas):
        inferred_type = ww.type_system.infer_logical_type(series)
        assert isinstance(inferred_type, expected_logical_types[index])


def test_updated_ltype_inference(integers, type_sys):
    inference_fn = type_sys.inference_functions[ww.logical_types.Integer]
    type_sys.remove_type(ww.logical_types.Integer)

    class Integer(LogicalType):
        primary_dtype = 'string'

    type_sys.add_type(Integer, inference_function=inference_fn)

    dtypes = ['int8', 'int16', 'int32', 'int64', 'intp', 'int', 'Int64']
    if _is_koalas_series(integers[0]):
        dtypes = get_koalas_dtypes(dtypes)

    for series in integers:
        for dtype in dtypes:
            inferred_type = type_sys.infer_logical_type(series.astype(dtype))
            assert isinstance(inferred_type, Integer)
            assert inferred_type.primary_dtype == 'string'
